from __future__ import annotations

import json
from pathlib import Path

from behave import then, when

from biblicus.corpus import Corpus
from biblicus.extraction_evaluation import ExtractionEvaluationItem, _resolve_item_id
from features.environment import run_biblicus


def _corpus_path(context, name: str) -> Path:
    return (context.workdir / name).resolve()


def _parse_json_output(standard_output: str) -> dict[str, object]:
    return json.loads(standard_output)


def _run_reference_from_context(context) -> str:
    extractor_id = context.last_extractor_id
    snapshot_id = context.last_extraction_snapshot_id
    assert isinstance(extractor_id, str) and extractor_id
    assert isinstance(snapshot_id, str) and snapshot_id
    return f"{extractor_id}:{snapshot_id}"


def _require_extraction_evaluation_output(context) -> dict[str, object]:
    if not hasattr(context, "last_extraction_evaluation"):
        result = getattr(context, "last_result", None)
        stderr = getattr(result, "stderr", "") if result is not None else ""
        raise AssertionError(f"Extraction evaluation output missing. stderr: {stderr}")
    return context.last_extraction_evaluation


@when("I attempt to resolve an extraction evaluation item without locators")
def step_attempt_resolve_missing_locators(context) -> None:
    item = ExtractionEvaluationItem.model_construct(
        item_id=None,
        source_uri=None,
        expected_text="",
        kind="gold",
    )
    try:
        _resolve_item_id(item, catalog_items={})
    except ValueError as exc:
        context.last_exception = exc
        return
    raise AssertionError("Expected resolve to raise a ValueError")


@when(
    'I attempt to resolve an extraction evaluation item with source uri "{source_uri}" using catalog from corpus "{corpus_name}"'
)
def step_attempt_resolve_source_uri(context, source_uri: str, corpus_name: str) -> None:
    corpus_root = _corpus_path(context, corpus_name)
    corpus = Corpus.open(corpus_root)
    catalog = corpus.load_catalog()
    item = ExtractionEvaluationItem.model_construct(
        item_id=None,
        source_uri=source_uri,
        expected_text="",
        kind="gold",
    )
    try:
        _resolve_item_id(item, catalog_items=catalog.items)
    except ValueError as exc:
        context.last_exception = exc
        return
    raise AssertionError("Expected resolve to raise a ValueError")


@then('the extraction evaluation resolver error mentions "{message}"')
def step_extraction_eval_resolver_error_mentions(context, message: str) -> None:
    exc = getattr(context, "last_exception", None)
    assert exc is not None
    assert message in str(exc)


@when('I create an extraction evaluation dataset "{filename}" with expected texts:')
def step_create_extraction_evaluation_dataset(context, filename: str) -> None:
    if not hasattr(context, "ingested_ids"):
        raise AssertionError("No ingested item identifiers recorded")
    expected_texts = [row["expected_text"] for row in context.table]
    item_ids = context.ingested_ids[-len(expected_texts) :]
    if len(item_ids) != len(expected_texts):
        raise AssertionError("Not enough ingested item identifiers to build dataset")
    items = [
        {
            "item_id": item_id,
            "expected_text": expected_text,
            "kind": "gold",
        }
        for item_id, expected_text in zip(item_ids, expected_texts)
    ]
    dataset = {
        "schema_version": 1,
        "name": "extraction-evaluation-dataset",
        "description": "Behavior-driven development extraction dataset",
        "items": items,
    }
    path = context.workdir / filename
    path.write_text(json.dumps(dataset, indent=2) + "\n", encoding="utf-8")


@when('I create an extraction evaluation dataset file "{filename}" with:')
def step_create_extraction_evaluation_dataset_file(context, filename: str) -> None:
    path = context.workdir / filename
    path.write_text(context.text, encoding="utf-8")


@when(
    'I create an extraction evaluation dataset "{filename}" for the last ingested item in corpus "{corpus_name}" '
    'using source uri and expected text "{expected_text}"'
)
def step_create_extraction_evaluation_dataset_with_source_uri(
    context, filename: str, corpus_name: str, expected_text: str
) -> None:
    if not getattr(context, "last_ingest", None):
        raise AssertionError("No last ingested item recorded")
    corpus_root = _corpus_path(context, corpus_name)
    corpus = Corpus.open(corpus_root)
    item_id = context.last_ingest["id"]
    source_uri = corpus.get_item(item_id).source_uri
    dataset = {
        "schema_version": 1,
        "name": "source-uri-dataset",
        "description": "Dataset keyed by source uniform resource identifier",
        "items": [
            {
                "source_uri": source_uri,
                "expected_text": expected_text,
                "kind": "gold",
            }
        ],
    }
    path = context.workdir / filename
    path.write_text(json.dumps(dataset, indent=2) + "\n", encoding="utf-8")


@when(
    'I evaluate extraction snapshot in corpus "{corpus_name}" using dataset "{filename}" '
    "and the latest extraction snapshot"
)
def step_evaluate_extraction_with_run(context, corpus_name: str, filename: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    run_ref = _run_reference_from_context(context)
    dataset_path = context.workdir / filename
    args = [
        "--corpus",
        str(corpus),
        "extract",
        "evaluate",
        "--snapshot",
        run_ref,
        "--dataset",
        str(dataset_path),
    ]
    result = run_biblicus(context, args)
    context.last_result = result
    if result.returncode == 0:
        context.last_extraction_evaluation = _parse_json_output(result.stdout)


@when('I evaluate extraction snapshot in corpus "{corpus_name}" using dataset "{filename}"')
def step_evaluate_extraction_without_run(context, corpus_name: str, filename: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    dataset_path = context.workdir / filename
    args = [
        "--corpus",
        str(corpus),
        "extract",
        "evaluate",
        "--dataset",
        str(dataset_path),
    ]
    result = run_biblicus(context, args)
    context.last_result = result
    if result.returncode == 0:
        context.last_extraction_evaluation = _parse_json_output(result.stdout)


@then("the extraction evaluation metrics include coverage_present {count:d}")
def step_extraction_eval_coverage_present(context, count: int) -> None:
    output = _require_extraction_evaluation_output(context)
    metrics = output["metrics"]
    assert metrics["coverage_present"] == float(count)


@then("the extraction evaluation metrics include coverage_empty {count:d}")
def step_extraction_eval_coverage_empty(context, count: int) -> None:
    output = _require_extraction_evaluation_output(context)
    metrics = output["metrics"]
    assert metrics["coverage_empty"] == float(count)


@then("the extraction evaluation metrics include coverage_missing {count:d}")
def step_extraction_eval_coverage_missing(context, count: int) -> None:
    output = _require_extraction_evaluation_output(context)
    metrics = output["metrics"]
    assert metrics["coverage_missing"] == float(count)


@then("the extraction evaluation metrics include processable_fraction {expected:g}")
def step_extraction_eval_processable_fraction(context, expected: float) -> None:
    output = _require_extraction_evaluation_output(context)
    metrics = output["metrics"]
    assert metrics["processable_fraction"] == expected


@then("the extraction evaluation metrics include average_similarity {expected:g}")
def step_extraction_eval_average_similarity(context, expected: float) -> None:
    output = _require_extraction_evaluation_output(context)
    metrics = output["metrics"]
    assert metrics["average_similarity"] == expected



from types import SimpleNamespace

from biblicus.evaluation.benchmark_runner import CategoryConfig, CategoryResult, run_category


def test_run_category_tracks_primary_metric(monkeypatch, tmp_path):
    cat_config = CategoryConfig(
        name="cat",
        dataset="ds",
        primary_metric="f1",
        pipelines=[SimpleNamespace(name="p1")],
    )
    cat_dir = tmp_path

    class FakeRunner:
        def run_pipeline(self, pipeline_name):
            return SimpleNamespace(
                avg_f1=0.9,
                avg_recall=0.8,
                avg_precision=0.85,
                avg_word_error_rate=0.1,
                avg_lcs_ratio=0.7,
                avg_bigram_overlap=0.6,
                avg_sequence_accuracy=0.5,
                total_documents=2,
            )

    result: CategoryResult = run_category(cat_dir, cat_config, FakeRunner())
    assert result.best_pipeline == "p1"
    assert result.primary_score == 0.9



"""
AWS Amplify Publisher for syncing Biblicus corpus data to cloud.
"""
import hashlib
import json
import os
import time
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional

import boto3
import requests


@dataclass
class SyncResult:
    """Result of a catalog sync operation."""

    skipped: bool = False
    reason: str = ""
    created: int = 0
    updated: int = 0
    deleted: int = 0
    errors: List[str] = field(default_factory=list)
    hash: str = ""


class AmplifyPublisher:
    """Publishes corpus events to AWS Amplify backend."""

    def __init__(self, corpus_name: str):
        self.corpus_name = corpus_name

        # Load config from environment or config file
        self._load_config()

        if not all([self.appsync_endpoint, self.api_key, self.s3_bucket]):
            raise ValueError(
                "Missing required Amplify configuration. "
                "Run 'biblicus dashboard configure' to set up credentials."
            )

        self.s3_client = boto3.client('s3', region_name=self.region)

    def _load_config(self):
        """Load configuration from environment or ~/.biblicus/amplify.env."""
        # Try environment variables first
        self.appsync_endpoint = os.getenv('AMPLIFY_APPSYNC_ENDPOINT')
        self.api_key = os.getenv('AMPLIFY_API_KEY')
        self.region = os.getenv('AWS_REGION', 'us-west-2')
        self.s3_bucket = os.getenv('AMPLIFY_S3_BUCKET')

        # If not in environment, try config file
        if not all([self.appsync_endpoint, self.api_key, self.s3_bucket]):
            config_path = Path.home() / '.biblicus' / 'amplify.env'
            if config_path.exists():
                for line in config_path.read_text().splitlines():
                    line = line.strip()
                    if line and not line.startswith('#') and '=' in line:
                        key, value = line.split('=', 1)
                        if key == 'AMPLIFY_APPSYNC_ENDPOINT' and not self.appsync_endpoint:
                            self.appsync_endpoint = value
                        elif key == 'AMPLIFY_API_KEY' and not self.api_key:
                            self.api_key = value
                        elif key == 'AWS_REGION':
                            self.region = value
                        elif key == 'AMPLIFY_S3_BUCKET':
                            if not self.s3_bucket:
                                self.s3_bucket = value

    def create_corpus(self) -> Dict[str, Any]:
        """Create or update corpus record."""
        mutation = """
        mutation CreateCorpus($input: CreateCorpusInput!) {
          createCorpus(input: $input) {
            name
            s3Prefix
            status
          }
        }
        """
        variables = {
            'input': {
                'name': self.corpus_name,
                's3Prefix': f'{self.corpus_name}',
                'status': 'IDLE',
            }
        }
        return self._execute_graphql(mutation, variables)

    def start_snapshot(
        self,
        snapshot_id: str,
        snapshot_type: str,
        total_items: int
    ) -> Dict[str, Any]:
        """Publish intent event when snapshot starts (before S3)."""
        mutation = """
        mutation CreateSnapshot($input: CreateSnapshotInput!) {
          createSnapshot(input: $input) {
            corpusId
            snapshotId
            status
            totalItems
          }
        }
        """
        variables = {
            'input': {
                'corpusId': self.corpus_name,
                'snapshotId': snapshot_id,
                'type': snapshot_type,
                'status': 'RUNNING',
                'totalItems': total_items,
                'completedItems': 0,
                'startTime': datetime.now(timezone.utc).isoformat(),
            }
        }
        return self._execute_graphql(mutation, variables)

    def update_snapshot_progress(
        self,
        snapshot_id: str,
        completed_items: int
    ) -> Dict[str, Any]:
        """Update snapshot progress during processing."""
        mutation = """
        mutation UpdateSnapshot($input: UpdateSnapshotInput!) {
          updateSnapshot(input: $input) {
            corpusId
            snapshotId
            completedItems
          }
        }
        """
        variables = {
            'input': {
                'corpusId': self.corpus_name,
                'snapshotId': snapshot_id,
                'completedItems': completed_items,
            }
        }
        return self._execute_graphql(mutation, variables)

    def complete_snapshot(
        self,
        snapshot_id: str,
        total_items: int,
        error_message: Optional[str] = None
    ) -> Dict[str, Any]:
        """Mark snapshot as completed or failed."""
        mutation = """
        mutation UpdateSnapshot($input: UpdateSnapshotInput!) {
          updateSnapshot(input: $input) {
            corpusId
            snapshotId
            status
          }
        }
        """
        variables = {
            'input': {
                'corpusId': self.corpus_name,
                'snapshotId': snapshot_id,
                'status': 'FAILED' if error_message else 'COMPLETED',
                'completedItems': total_items,
                'endTime': datetime.now(timezone.utc).isoformat(),
            }
        }
        if error_message:
            variables['input']['errorMessage'] = error_message

        return self._execute_graphql(mutation, variables)

    def register_file(
        self,
        file_path: str,
        relative_to: Path
    ) -> Dict[str, Any]:
        """Register file as LOCAL_ONLY before upload."""
        relative_path = str(Path(file_path).relative_to(relative_to))

        mutation = """
        mutation CreateFileMetadata($input: CreateFileMetadataInput!) {
          createFileMetadata(input: $input) {
            corpusId
            filePath
            status
          }
        }
        """
        variables = {
            'input': {
                'corpusId': self.corpus_name,
                'filePath': relative_path,
                'status': 'LOCAL_ONLY',
            }
        }
        return self._execute_graphql(mutation, variables)

    def upload_file(
        self,
        local_path: Path,
        relative_to: Path
    ) -> Dict[str, Any]:
        """Upload file to S3 (direct mirror of local structure)."""
        relative_path = str(local_path.relative_to(relative_to))
        s3_key = f'{self.corpus_name}/{relative_path}'

        # Update status to UPLOADING
        mutation = """
        mutation UpdateFileMetadata($input: UpdateFileMetadataInput!) {
          updateFileMetadata(input: $input) {
            corpusId
            filePath
            status
          }
        }
        """
        variables = {
            'input': {
                'corpusId': self.corpus_name,
                'filePath': relative_path,
                'status': 'UPLOADING',
            }
        }
        self._execute_graphql(mutation, variables)

        # Upload to S3 (this will trigger Lambda → AVAILABLE)
        with open(local_path, 'rb') as f:
            self.s3_client.put_object(
                Bucket=self.s3_bucket,
                Key=s3_key,
                Body=f
            )

        return {'s3Key': s3_key}

    def sync_catalog(self, catalog_path: Path, force: bool = False) -> SyncResult:
        """Sync catalog.json to Amplify (idempotent)."""
        from biblicus.models import CorpusCatalog

        # 1. Load and hash catalog
        catalog = CorpusCatalog.model_validate_json(catalog_path.read_text())
        catalog_hash = self._compute_catalog_hash(catalog)

        # 2. Check if changed (idempotency)
        try:
            metadata = self._get_catalog_metadata()
            if not force and metadata and metadata.get('catalogHash') == catalog_hash:
                return SyncResult(skipped=True, reason="Catalog unchanged", hash=catalog_hash)
        except Exception:
            metadata = None

        # 3. Choose sync strategy
        if len(catalog.items) < 100 or force:
            result = self._full_replacement_sync(catalog, catalog_hash)
        else:
            result = self._incremental_sync(catalog, metadata, catalog_hash)

        # 4. Update metadata
        self._sync_catalog_metadata(catalog, catalog_hash)

        return result

    def _compute_catalog_hash(self, catalog) -> str:
        """Hash catalog excluding timestamps for idempotency."""
        items_data = sorted([
            (item.id, item.sha256, item.relpath)
            for item in catalog.items.values()
        ])
        return hashlib.sha256(json.dumps(items_data).encode()).hexdigest()

    def _get_catalog_metadata(self) -> Optional[Dict[str, Any]]:
        """Get current catalog metadata from Amplify."""
        query = """
        query GetCatalogMetadata($corpusId: ID!) {
          getCatalogMetadata(corpusId: $corpusId) {
            corpusId
            catalogHash
            itemCount
            lastSyncedAt
          }
        }
        """
        variables = {'corpusId': self.corpus_name}
        try:
            result = self._execute_graphql(query, variables)
            return result.get('getCatalogMetadata')
        except Exception:
            return None

    def _sync_catalog_metadata(self, catalog, catalog_hash: str):
        """Update catalog metadata in Amplify."""
        mutation = """
        mutation UpdateCatalogMetadata($input: UpdateCatalogMetadataInput!) {
          updateCatalogMetadata(input: $input) {
            corpusId
            catalogHash
          }
        }
        """
        variables = {
            'input': {
                'corpusId': self.corpus_name,
                'catalogHash': catalog_hash,
                'itemCount': len(catalog.items),
                'lastSyncedAt': datetime.now(timezone.utc).isoformat(),
                'schemaVersion': catalog.schema_version,
                'configurationId': catalog.configuration_id if hasattr(catalog, 'configuration_id') else None,
                'corpusUri': catalog.corpus_uri if hasattr(catalog, 'corpus_uri') else None,
            }
        }
        try:
            self._execute_graphql(mutation, variables)
        except Exception:
            # Try create if update failed
            mutation = mutation.replace('updateCatalogMetadata', 'createCatalogMetadata')
            mutation = mutation.replace('UpdateCatalogMetadataInput', 'CreateCatalogMetadataInput')
            self._execute_graphql(mutation, variables)

    def _full_replacement_sync(self, catalog, catalog_hash: str) -> SyncResult:
        """Replace all catalog items (for small catalogs)."""
        errors = []
        created = 0

        # Create all items
        for item_id, item in catalog.items.items():
            try:
                self._create_catalog_item(item)
                created += 1
            except Exception as e:
                errors.append(f"Failed to sync item {item.id}: {e}")

        return SyncResult(
            skipped=False,
            created=created,
            updated=0,
            deleted=0,
            errors=errors,
            hash=catalog_hash
        )

    def _incremental_sync(self, catalog, metadata: Optional[Dict], catalog_hash: str) -> SyncResult:
        """Sync only changed items (for large catalogs)."""
        # For MVP, just do full replacement
        # TODO: Implement proper diff logic
        return self._full_replacement_sync(catalog, catalog_hash)

    def _retry_attempts(self) -> Iterable[int]:
        return range(3)

    def _create_catalog_item(self, item):
        """Create or update a catalog item with retry."""
        mutation = """
        mutation CreateCatalogItem($input: CreateCatalogItemInput!) {
          createCatalogItem(input: $input) {
            corpusId
            itemId
          }
        }
        """

        variables = {
            'input': {
                'corpusId': self.corpus_name,
                'itemId': item.id,
                'relpath': item.relpath,
                'sha256': item.sha256,
                'bytes': item.bytes,
                'mediaType': item.media_type,
                'title': item.title if hasattr(item, 'title') else None,
                'tags': item.tags if hasattr(item, 'tags') else [],
                'metadataJson': json.dumps(item.metadata) if hasattr(item, 'metadata') and item.metadata else None,
                'createdAt': datetime.now(timezone.utc).isoformat(),
                'sourceUri': item.source_uri if hasattr(item, 'source_uri') else None,
                'hasExtraction': False,
            }
        }

        published = False
        last_error: Exception | None = None
        for attempt in self._retry_attempts():
            try:
                self._execute_graphql(mutation, variables)
                published = True
                break
            except Exception as e:
                last_error = e
                if attempt < 2 and 'Network' in str(e):
                    time.sleep(2 ** attempt)
                    continue
                break
        if published:
            return
        if last_error is not None:
            raise last_error

    def _execute_graphql(self, query: str, variables: Dict[str, Any]) -> Dict[str, Any]:
        """Execute GraphQL mutation against AppSync."""
        headers = {
            'Content-Type': 'application/json',
            'x-api-key': self.api_key,
        }

        payload = {
            'query': query,
            'variables': variables,
        }

        response = requests.post(
            self.appsync_endpoint,
            headers=headers,
            json=payload,
            timeout=10
        )
        response.raise_for_status()

        result = response.json()
        if 'errors' in result:
            raise Exception(f"GraphQL errors: {result['errors']}")

        return result.get('data', {})



from types import SimpleNamespace

from biblicus.analysis.topic_modeling import _remove_entities_from_text


def test_remove_entities_skips_invalid_spans():
    text = "hello world"
    bad_entity = SimpleNamespace(label_="PERSON", start_char=5, end_char=5)
    result = _remove_entities_from_text(
        text=text,
        entities=[bad_entity],
        entity_types={"PERSON"},
        replace_with="[REMOVED]",
    )
    assert result == text



#!/usr/bin/env python3
"""
Configure Amplify integration from amplify_outputs.json.

Usage:
    python scripts/configure_amplify_from_outputs.py
"""

import json
import os
from pathlib import Path

def main():
    # Load amplify_outputs.json
    outputs_path = Path(__file__).parent.parent / 'apps' / 'dashboard' / 'amplify_outputs.json'

    if not outputs_path.exists():
        print(f'Error: {outputs_path} not found')
        print('Run "npx ampx sandbox" in apps/dashboard first')
        return 1

    with open(outputs_path) as f:
        outputs = json.load(f)

    # Extract values
    endpoint = outputs['data']['url']
    api_key = outputs['data']['api_key']
    bucket = outputs['storage']['bucket_name']
    region = outputs['data']['aws_region']

    # Create config directory
    config_dir = Path.home() / '.biblicus'
    config_dir.mkdir(exist_ok=True)

    # Write config
    config_path = config_dir / 'amplify.env'
    with open(config_path, 'w') as f:
        f.write(f'AMPLIFY_APPSYNC_ENDPOINT={endpoint}\n')
        f.write(f'AMPLIFY_API_KEY={api_key}\n')
        f.write(f'AMPLIFY_S3_BUCKET={bucket}\n')
        f.write(f'AWS_REGION={region}\n')

    print(f'✓ Configuration saved to {config_path}')
    print(f'\nEndpoint: {endpoint}')
    print(f'API Key: {api_key[:10]}...')
    print(f'Bucket: {bucket}')
    print(f'Region: {region}')
    print(f'\nTo enable auto-sync after extraction, run:')
    print(f'  export AMPLIFY_AUTO_SYNC_CATALOG=true')

    return 0

if __name__ == '__main__':
    exit(main())



"""
Download the AG News dataset into a Biblicus corpus.
"""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Dict

from biblicus.corpus import Corpus

LABELS = ["World", "Sports", "Business", "Sci/Tech"]


def _prepare_corpus(path: Path, *, force: bool) -> Corpus:
    """
    Initialize or open a corpus for dataset downloads.

    :param path: Corpus path.
    :type path: Path
    :param force: Whether to purge existing corpus content.
    :type force: bool
    :return: Corpus instance.
    :rtype: Corpus
    :raises ValueError: If the target path is non-empty without force.
    """
    if (path / ".biblicus" / "config.json").is_file():
        corpus = Corpus.open(path)
        if force:
            corpus.purge(confirm=corpus.name)
        return corpus
    if path.exists() and any(path.iterdir()) and not force:
        raise ValueError("Target corpus directory is not empty. Use --force to initialize anyway.")
    return Corpus.init(path, force=True)


def _count_existing_ag_news_items(corpus: Corpus) -> int:
    """
    Count existing AG News items in the corpus.

    :param corpus: Corpus instance to inspect.
    :type corpus: Corpus
    :return: Count of existing AG News items.
    :rtype: int
    """
    catalog = corpus.load_catalog()
    return sum(1 for item in catalog.items.values() if "ag_news" in item.tags)


def download_ag_news_corpus(
    *, corpus_path: Path, split: str, limit: int, force: bool, resume: bool
) -> Dict[str, int]:
    """
    Download AG News samples into a corpus.

    :param corpus_path: Corpus path to create or reuse.
    :type corpus_path: Path
    :param split: Dataset split to load.
    :type split: str
    :param limit: Maximum number of samples to ingest.
    :type limit: int
    :param force: Whether to purge existing corpus content.
    :type force: bool
    :return: Ingestion statistics.
    :rtype: dict[str, int]
    """
    try:
        from datasets import load_dataset
    except ImportError as exc:
        raise RuntimeError(
            "AG News download requires the datasets library. "
            "Install with: python3 -m pip install 'biblicus[datasets]'"
        ) from exc

    corpus = _prepare_corpus(corpus_path, force=force)
    existing = 0
    if resume and not force:
        existing = _count_existing_ag_news_items(corpus)
        if existing >= limit:
            return {"ingested": 0, "skipped": 0, "existing": existing, "target": limit}
    dataset = load_dataset("ag_news", split=split)
    ingested = 0
    skipped = 0
    remaining = max(0, limit - existing)
    for index, record in enumerate(dataset):
        if index < existing:
            continue
        if ingested >= remaining:
            break
        text = (record.get("text") or "").strip()
        if not text:
            skipped += 1
            continue
        label_value = record.get("label")
        label_name = LABELS[label_value] if isinstance(label_value, int) else "unknown"
        title = f"AG News {split} {index}"
        tags = ["ag_news", f"label:{label_name}"]
        corpus.ingest_note(text, title=title, tags=tags)
        ingested += 1
    corpus.reindex()
    return {"ingested": ingested, "skipped": skipped, "existing": existing, "target": limit}


def build_parser() -> argparse.ArgumentParser:
    """
    Build the command-line interface argument parser.

    :return: Argument parser.
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(description="Download AG News into Biblicus.")
    parser.add_argument("--corpus", required=True, help="Corpus path to initialize or reuse.")
    parser.add_argument("--split", default="train", help="Dataset split to ingest.")
    parser.add_argument("--limit", type=int, default=1000, help="Number of rows to ingest.")
    parser.add_argument(
        "--force", action="store_true", help="Initialize even if the directory is not empty."
    )
    parser.add_argument(
        "--resume",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Skip items already ingested into the corpus.",
    )
    return parser


def main() -> int:
    """
    Entry point for the AG News download script.

    :return: Exit code.
    :rtype: int
    """
    parser = build_parser()
    args = parser.parse_args()
    stats = download_ag_news_corpus(
        corpus_path=Path(args.corpus).resolve(),
        split=args.split,
        limit=args.limit,
        force=args.force,
        resume=args.resume,
    )
    print(json.dumps(stats, indent=2))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""
Unit tests for tool loop safeguards and feedback.
"""

from __future__ import annotations

import json
import unittest

from biblicus.ai.llm import ChatCompletionResult
from biblicus.ai.models import AiProvider, LlmClientConfig
from biblicus.text import tool_loop as tool_loop_module
from biblicus.text.tool_loop import apply_unique_str_replace, run_tool_loop


class FakeNoOpCompletion:
    """
    Fake completion that attempts a no-op replacement.
    """

    def __call__(self, **_kwargs: object) -> ChatCompletionResult:
        """
        Return a synthetic no-op completion result.
        """
        tool_calls = [
            {
                "id": "call_1",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps({"old_str": "Hello", "new_str": "Hello"}),
                },
            },
            {"id": "call_2", "type": "function", "function": {"name": "done", "arguments": "{}"}},
        ]
        return ChatCompletionResult(text="", tool_calls=tool_calls)


class FakeMultiMatchCompletion:
    """
    Fake completion that uses a short old_str with multiple matches.
    """

    def __call__(self, **_kwargs: object) -> ChatCompletionResult:
        """
        Return a synthetic completion with ambiguous matches.
        """
        tool_calls = [
            {
                "id": "call_1",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps({"old_str": "hi", "new_str": "<span>hi</span>"}),
                },
            },
            {"id": "call_2", "type": "function", "function": {"name": "done", "arguments": "{}"}},
        ]
        return ChatCompletionResult(text="", tool_calls=tool_calls)


class FakeRemovalCompletion:
    """
    Fake completion that attempts to remove content.
    """

    def __call__(self, **_kwargs: object) -> ChatCompletionResult:
        """
        Return a synthetic completion that removes content.
        """
        tool_calls = [
            {
                "id": "call_1",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps({"old_str": "Hello world", "new_str": "Hello"}),
                },
            },
            {"id": "call_2", "type": "function", "function": {"name": "done", "arguments": "{}"}},
        ]
        return ChatCompletionResult(text="", tool_calls=tool_calls)


class FakeShortNotFoundCompletion:
    """
    Fake completion that uses a short old_str with no matches.
    """

    def __call__(self, **_kwargs: object) -> ChatCompletionResult:
        """
        Return a synthetic completion that yields no matches.
        """
        tool_calls = [
            {
                "id": "call_1",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps({"old_str": "a", "new_str": "<span>a</span>"}),
                },
            },
            {"id": "call_2", "type": "function", "function": {"name": "done", "arguments": "{}"}},
        ]
        return ChatCompletionResult(text="", tool_calls=tool_calls)


class TestToolLoopSafeguards(unittest.TestCase):
    """
    Unit tests for tool loop safeguards.
    """

    def _run(self, *, text: str, fake_completion: object) -> tool_loop_module.ToolLoopResult:
        """
        Run the tool loop with a fake completion implementation.
        """
        original = tool_loop_module.chat_completion
        tool_loop_module.chat_completion = fake_completion
        try:
            return run_tool_loop(
                text=text,
                client=LlmClientConfig(
                    provider=AiProvider.OPENAI,
                    model="gpt-4o-mini",
                    api_key="test-openai-key",
                    response_format="json_object",
                ),
                system_prompt="You are a virtual file editor.\nCurrent text:\n---\n{text}\n---",
                prompt_template="Return the requested text.",
                max_rounds=1,
                max_edits_per_round=5,
                apply_str_replace=apply_unique_str_replace,
            )
        finally:
            tool_loop_module.chat_completion = original

    def test_no_op_replacement_is_rejected(self) -> None:
        """
        Ensure no-op replacements are rejected.
        """
        result = self._run(text="Hello", fake_completion=FakeNoOpCompletion())
        self.assertIn("requires str_replace to make a change", result.last_error or "")
        self.assertFalse(result.done)
        self.assertIn("Your last tool call failed", result.messages[-1]["content"])

    def test_multiple_matches_are_rejected_with_guidance(self) -> None:
        """
        Ensure ambiguous replacements are rejected with guidance.
        """
        result = self._run(text="hi hi", fake_completion=FakeMultiMatchCompletion())
        self.assertIn("found 2 matches", result.last_error or "")
        self.assertIn("longer unique old_str", result.messages[-1]["content"])

    def test_removal_is_rejected(self) -> None:
        """
        Ensure replacements that remove content are rejected.
        """
        result = self._run(text="Hello world", fake_completion=FakeRemovalCompletion())
        self.assertIn("may only insert markup tags", result.last_error or "")

    def test_short_old_str_hints_view(self) -> None:
        """
        Ensure short unmatched strings produce a helpful hint.
        """
        result = self._run(text="bbb", fake_completion=FakeShortNotFoundCompletion())
        self.assertIn("found 0 matches", result.last_error or "")
        self.assertIn("call view", result.messages[-1]["content"])


if __name__ == "__main__":
    unittest.main()



import types

from biblicus.extractors import deepgram_stt


def test_deepgram_response_to_dict_paths():
    class HasToDict:
        def to_dict(self):
            return {"a": 1}

    class HasToJson:
        def to_json(self):
            return '{"b":2}'

    class HasModelDump:
        def model_dump(self):
            return {"c": 3}

    class HasDict:
        def dict(self):
            return {"d": 4}

    assert deepgram_stt._deepgram_response_to_dict(HasToDict())["a"] == 1
    assert deepgram_stt._deepgram_response_to_dict(HasToJson())["b"] == 2
    assert deepgram_stt._deepgram_response_to_dict(HasModelDump())["c"] == 3
    assert deepgram_stt._deepgram_response_to_dict(HasDict())["d"] == 4
    assert deepgram_stt._deepgram_response_to_dict(types.SimpleNamespace(e=5)) == {"e": 5}


def test_normalize_deepgram_value_complex():
    nested = {"ts": deepgram_stt.datetime(2020, 1, 1), "list": [types.SimpleNamespace(x=1)]}
    result = deepgram_stt._normalize_deepgram_value(nested)
    assert "ts" in result and "list" in result



from __future__ import annotations

import json
import subprocess
import sys
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional

from behave import given, then, when

from biblicus.corpus import Corpus
from biblicus.evaluation import _snapshot_artifact_bytes
from biblicus.models import Evidence, QueryBudget
from biblicus.retrieval import apply_budget
from biblicus.retrievers.scan import _build_snippet, _find_first_match
from biblicus.retrievers.sqlite_full_text_search import (
    SqliteFullTextSearchConfiguration,
    _build_full_text_search_index,
    _iter_chunks,
)
from features.environment import run_biblicus


def _corpus_path(context, name: str) -> Path:
    return (context.workdir / name).resolve()


def _parse_json_output(standard_output: str) -> Dict[str, Any]:
    return json.loads(standard_output)


def _table_key_value(row) -> tuple[str, str]:
    if "key" in row.headings and "value" in row.headings:
        return row["key"].strip(), row["value"].strip()
    return row[0].strip(), row[1].strip()


def _parse_token_list(raw: str) -> List[str]:
    return [token for token in raw.split(",")]


def _parse_span(raw: str) -> Optional[tuple[int, int]]:
    if raw.lower() == "none":
        return None
    start, end = raw.split("..", 1)
    return int(start), int(end)


def _normalize_empty_text(text: str) -> str:
    return "" if text == "<empty>" else text


def _build_evidence_items(texts: Iterable[str]) -> List[Evidence]:
    evidence_items: List[Evidence] = []
    for index, text in enumerate(texts, start=1):
        evidence_items.append(
            Evidence(
                item_id=f"item-{index}",
                source_uri="source",
                media_type="text/plain",
                score=float(len(text)),
                rank=index,
                text=text,
                content_ref=None,
                span_start=None,
                span_end=None,
                stage="scan",
                configuration_id="configuration",
                snapshot_id="snapshot",
                hash=None,
            )
        )
    return evidence_items


def _evidence_keys_from_query(
    query_json: Dict[str, Any],
) -> set[tuple[object, object, object, object]]:
    evidence = query_json.get("evidence") or []
    keys: set[tuple[object, object, object, object]] = set()
    for item in evidence:
        keys.add(
            (
                item.get("item_id"),
                item.get("span_start"),
                item.get("span_end"),
                item.get("hash"),
            )
        )
    return keys


@when('I build a "{retriever_id}" retrieval snapshot in corpus "{corpus_name}"')
def step_build_snapshot(context, retriever_id: str, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    result = run_biblicus(
        context,
        [
            "--corpus",
            str(corpus),
            "build",
            "--retriever",
            retriever_id,
            "--configuration-name",
            "default",
        ],
    )
    assert result.returncode == 0, result.stderr
    context.last_snapshot = _parse_json_output(result.stdout)
    context.last_snapshot_id = context.last_snapshot.get("snapshot_id")


@when('I build a "{retriever_id}" retrieval snapshot in corpus "{corpus_name}" with config:')
def step_build_snapshot_with_config(context, retriever_id: str, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    args = [
        "--corpus",
        str(corpus),
        "build",
        "--retriever",
        retriever_id,
        "--configuration-name",
        "default",
    ]
    for row in context.table:
        key, value = _table_key_value(row)
        args.extend(["--override", f"{key}={value}"])
    result = run_biblicus(context, args)
    assert result.returncode == 0, result.stderr
    context.last_snapshot = _parse_json_output(result.stdout)
    context.last_snapshot_id = context.last_snapshot.get("snapshot_id")


@when(
    'I attempt to build a "{retriever_id}" retrieval snapshot in corpus "{corpus_name}" with config:'
)
def step_attempt_build_snapshot_with_config(context, retriever_id: str, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    args = [
        "--corpus",
        str(corpus),
        "build",
        "--retriever",
        retriever_id,
        "--configuration-name",
        "default",
    ]
    for row in context.table:
        key, value = _table_key_value(row)
        args.extend(["--override", f"{key}={value}"])
    context.last_result = run_biblicus(context, args)


@when('I query with the latest snapshot for "{query_text}" and budget:')
def step_query_latest_snapshot(context, query_text: str) -> None:
    assert context.last_snapshot_id
    corpus = _corpus_path(context, "corpus")
    args = [
        "--corpus",
        str(corpus),
        "query",
        "--snapshot",
        context.last_snapshot_id,
        "--query",
        query_text,
    ]
    for row in context.table:
        key, value = _table_key_value(row)
        args.extend([f"--{key.replace('_', '-')}", value])
    result = run_biblicus(context, args)
    assert result.returncode == 0, result.stderr
    context.last_query = _parse_json_output(result.stdout)


@when('I attempt to query with the latest snapshot for "{query_text}" and budget:')
def step_attempt_query_latest_snapshot(context, query_text: str) -> None:
    assert context.last_snapshot_id
    corpus = _corpus_path(context, "corpus")
    args = [
        "--corpus",
        str(corpus),
        "query",
        "--snapshot",
        context.last_snapshot_id,
        "--query",
        query_text,
    ]
    for row in context.table:
        key, value = _table_key_value(row)
        args.extend([f"--{key.replace('_', '-')}", value])
    context.last_result = run_biblicus(context, args)


@when("I remember the query evidence keys")
def step_remember_query_evidence_keys(context) -> None:
    assert context.last_query is not None
    context.remembered_evidence_keys = _evidence_keys_from_query(context.last_query)


@then("the query evidence keys are disjoint from the remembered keys")
def step_query_evidence_keys_disjoint(context) -> None:
    assert context.last_query is not None
    assert hasattr(context, "remembered_evidence_keys")
    current_keys = _evidence_keys_from_query(context.last_query)
    overlap = current_keys.intersection(context.remembered_evidence_keys)
    assert not overlap


@when("I attempt to query the latest snapshot with an invalid budget")
def step_query_latest_snapshot_invalid_budget(context) -> None:
    assert context.last_snapshot_id
    corpus = _corpus_path(context, "corpus")
    args = [
        "--corpus",
        str(corpus),
        "query",
        "--snapshot",
        context.last_snapshot_id,
        "--query",
        "test",
        "--max-total-items",
        "0",
        "--maximum-total-characters",
        "10",
        "--max-items-per-source",
        "1",
    ]
    run_biblicus(context, args)


@then('the query returns evidence with stage "{stage}"')
def step_query_returns_stage(context, stage: str) -> None:
    evidence = context.last_query.get("evidence") or []
    assert evidence
    assert all(item.get("stage") == stage for item in evidence)


@then("the query evidence count is {count:d}")
def step_query_evidence_count(context, count: int) -> None:
    evidence = context.last_query.get("evidence") or []
    assert len(evidence) == count


@then("the query evidence includes the last ingested item identifier")
def step_query_includes_last_ingested(context) -> None:
    assert context.last_ingest is not None
    evidence = context.last_query.get("evidence") or []
    ids = {item.get("item_id") for item in evidence}
    assert context.last_ingest["id"] in ids


@then("the latest snapshot stats include text_items {count:d}")
def step_run_stats_text_items(context, count: int) -> None:
    stats = context.last_snapshot.get("stats") or {}
    assert stats.get("text_items") == count


@then("the latest snapshot stats include chunks {count:d}")
def step_run_stats_chunks(context, count: int) -> None:
    stats = context.last_snapshot.get("stats") or {}
    assert stats.get("chunks") == count


@when('I create an evaluation dataset at "{filename}" with queries:')
def step_create_eval_dataset(context, filename: str) -> None:
    assert context.ingested_ids
    last_id = context.ingested_ids[-1]
    previous_id = (
        context.ingested_ids[-2] if len(context.ingested_ids) > 1 else context.ingested_ids[-1]
    )
    queries = []
    for idx, row in enumerate(context.table, start=1):
        query_text = row["query_text"] if "query_text" in row.headings else row[0]
        expected_key = row["expected_item"] if "expected_item" in row.headings else row[1]
        if expected_key == "last_ingested":
            expected_id = last_id
        elif expected_key == "previous_item":
            expected_id = previous_id
        else:
            expected_id = expected_key
        queries.append(
            {
                "query_id": f"q{idx}",
                "query_text": query_text,
                "expected_item_id": expected_id,
                "kind": "gold",
            }
        )
    dataset = {
        "schema_version": 1,
        "name": "behavior-driven-development-dataset",
        "description": "Behavior-driven development evaluation dataset",
        "queries": queries,
    }
    path = context.workdir / filename
    path.write_text(json.dumps(dataset, indent=2) + "\n", encoding="utf-8")
    context.dataset_path = path


@when('I evaluate the latest snapshot with dataset "{filename}" and budget:')
def step_eval_latest_run(context, filename: str) -> None:
    assert context.last_snapshot_id
    path = context.workdir / filename
    args = [
        "--corpus",
        str(_corpus_path(context, "corpus")),
        "eval",
        "--snapshot",
        context.last_snapshot_id,
        "--dataset",
        str(path),
    ]
    for row in context.table:
        key, value = _table_key_value(row)
        args.extend([f"--{key.replace('_', '-')}", value])
    result = run_biblicus(context, args)
    assert result.returncode == 0, result.stderr
    context.last_eval = _parse_json_output(result.stdout)


@then("the evaluation reports mean reciprocal rank {expected:g}")
def step_eval_mean_reciprocal_rank(context, expected: float) -> None:
    metrics = context.last_eval.get("metrics") or {}
    assert metrics.get("mean_reciprocal_rank") == expected


@then("the evaluation reports hit_rate {expected:g}")
def step_eval_hit_rate(context, expected: float) -> None:
    metrics = context.last_eval.get("metrics") or {}
    assert metrics.get("hit_rate") == expected


@then("the evaluation system reports index_bytes greater than 0")
def step_eval_index_bytes(context) -> None:
    system = context.last_eval.get("system") or {}
    assert system.get("index_bytes", 0) > 0


@then("the evaluation system reports index_bytes {expected:g}")
def step_eval_index_bytes_value(context, expected: float) -> None:
    system = context.last_eval.get("system") or {}
    assert system.get("index_bytes") == expected


@when(
    'I create a source uniform resource identifier evaluation dataset at "{filename}" '
    'for query "{query_text}"'
)
def step_create_source_uri_dataset(context, filename: str, query_text: str) -> None:
    source_uri = "text"
    last_ingest = getattr(context, "last_ingest", None)
    if last_ingest is not None:
        item_id = last_ingest.get("id")
        if isinstance(item_id, str):
            corpus = Corpus.open(_corpus_path(context, "corpus"))
            source_uri = corpus.get_item(item_id).source_uri
    dataset = {
        "schema_version": 1,
        "name": "source-uniform-resource-identifier-dataset",
        "description": "Dataset keyed by source uniform resource identifier",
        "queries": [
            {
                "query_id": "q1",
                "query_text": query_text,
                "expected_source_uri": source_uri,
                "kind": "gold",
            }
        ],
    }
    path = context.workdir / filename
    path.write_text(json.dumps(dataset, indent=2) + "\n", encoding="utf-8")
    context.dataset_path = path


@given('I create an empty evaluation dataset at "{filename}"')
@when('I create an empty evaluation dataset at "{filename}"')
def step_create_empty_dataset(context, filename: str) -> None:
    dataset = {
        "schema_version": 1,
        "name": "empty-dataset",
        "description": "Empty evaluation dataset",
        "queries": [],
    }
    path = context.workdir / filename
    path.write_text(json.dumps(dataset, indent=2) + "\n", encoding="utf-8")
    context.dataset_path = path


@when("I delete the latest snapshot artifacts")
def step_delete_latest_snapshot_artifacts(context) -> None:
    assert context.last_snapshot_id
    corpus = Corpus.open(_corpus_path(context, "corpus"))
    snapshot = corpus.load_snapshot(context.last_snapshot_id)
    for relpath in snapshot.snapshot_artifacts:
        path = corpus.root / relpath
        if path.exists():
            path.unlink()


@when("I measure the latest snapshot artifact bytes")
def step_measure_latest_snapshot_artifacts(context) -> None:
    assert context.last_snapshot_id
    corpus = Corpus.open(_corpus_path(context, "corpus"))
    snapshot = corpus.load_snapshot(context.last_snapshot_id)
    context.latest_snapshot_bytes = _snapshot_artifact_bytes(corpus, snapshot)


@then("the snapshot artifact bytes are {count:d}")
def step_snapshot_artifact_bytes(context, count: int) -> None:
    assert context.latest_snapshot_bytes == count


@then("the snapshot artifact bytes are greater than 0")
def step_snapshot_artifact_bytes_positive(context) -> None:
    assert int(getattr(context, "latest_snapshot_bytes", 0)) > 0


@then("the query evidence includes span offsets")
def step_query_evidence_includes_span_offsets(context) -> None:
    evidence = context.last_query.get("evidence") or []
    assert evidence
    for item in evidence:
        assert item.get("span_start") is not None
        assert item.get("span_end") is not None


@when('I attempt to query the latest snapshot with retriever "{retriever_id}"')
def step_query_latest_snapshot_retriever_mismatch(context, retriever_id: str) -> None:
    assert context.last_snapshot_id
    corpus = _corpus_path(context, "corpus")
    args = [
        "--corpus",
        str(corpus),
        "query",
        "--snapshot",
        context.last_snapshot_id,
        "--retriever",
        retriever_id,
        "--query",
        "test",
    ]
    context.last_result = run_biblicus(context, args)


@when('I compute a scan match span for text "{text}" and tokens "{tokens}"')
def step_compute_scan_span(context, text: str, tokens: str) -> None:
    token_list = _parse_token_list(tokens)
    context.scan_span = _find_first_match(text, token_list)


@then('the scan match span is "{span}"')
def step_scan_span_is(context, span: str) -> None:
    expected_span = _parse_span(span)
    assert context.scan_span == expected_span


@when('I build a scan snippet from text "{text}" with span "{span}" and max chars {max_chars:d}')
def step_build_scan_snippet(context, text: str, span: str, max_chars: int) -> None:
    parsed_span = _parse_span(span)
    normalized_text = _normalize_empty_text(text)
    context.scan_snippet = _build_snippet(normalized_text, parsed_span, max_chars=max_chars)


@then('the scan snippet is "{snippet}"')
def step_scan_snippet_is(context, snippet: str) -> None:
    expected = _normalize_empty_text(snippet)
    assert context.scan_snippet == expected


@when(
    'I split text "{text}" into sqlite chunks with size {chunk_size:d} and overlap {chunk_overlap:d}'
)
def step_split_sqlite_chunks(context, text: str, chunk_size: int, chunk_overlap: int) -> None:
    normalized_text = _normalize_empty_text(text)
    context.sqlite_chunks = list(
        _iter_chunks(normalized_text, chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    )


@then("the sqlite chunk count is {count:d}")
def step_sqlite_chunk_count(context, count: int) -> None:
    assert len(context.sqlite_chunks) == count


@when('I rebuild a SQLite full-text search index for corpus "{corpus_name}" at "{relative_path}"')
def step_rebuild_sqlite_full_text_search(context, corpus_name: str, relative_path: str) -> None:
    corpus = Corpus.open(_corpus_path(context, corpus_name))
    db_path = corpus.root / relative_path
    db_path.parent.mkdir(parents=True, exist_ok=True)
    db_path.write_text("stale", encoding="utf-8")
    config = SqliteFullTextSearchConfiguration(
        chunk_size=5, chunk_overlap=2, snippet_characters=120
    )
    stats = _build_full_text_search_index(
        db_path=db_path,
        corpus=corpus,
        items=corpus.load_catalog().items.values(),
        configuration=config,
        extraction_reference=None,
    )
    context.sqlite_full_text_search_path = db_path
    context.sqlite_full_text_search_stats = stats


@then("the SQLite full-text search index file exists")
def step_sqlite_full_text_search_index_exists(context) -> None:
    assert context.sqlite_full_text_search_path.exists()


@when("I apply a budget with no per-source or character limits")
def step_apply_budget_without_limits(context) -> None:
    evidence = _build_evidence_items(["alpha", "beta", "gamma"])
    budget = QueryBudget(
        max_total_items=2, maximum_total_characters=None, max_items_per_source=None
    )
    context.budgeted_evidence = apply_budget(evidence, budget)


@when("I apply a budget with a per-source limit of {limit:d}")
def step_apply_budget_with_per_source_limit(context, limit: int) -> None:
    evidence = _build_evidence_items(["alpha", "beta", "gamma"])
    budget = QueryBudget(
        max_total_items=5, maximum_total_characters=None, max_items_per_source=limit
    )
    context.budgeted_evidence = apply_budget(evidence, budget)


@then("the budget returns {count:d} evidence items")
def step_budget_count(context, count: int) -> None:
    assert len(context.budgeted_evidence) == count


@then('the budget returns evidence ranks "{ranks}"')
def step_budget_ranks(context, ranks: str) -> None:
    expected = [int(value) for value in ranks.split(",")]
    observed = [item.rank for item in context.budgeted_evidence]
    assert observed == expected


@when('I create an invalid evaluation dataset at "{filename}" for query "{query_text}"')
def step_create_invalid_dataset(context, filename: str, query_text: str) -> None:
    dataset = {
        "schema_version": 1,
        "name": "invalid-dataset",
        "description": "Missing expectation fields",
        "queries": [
            {
                "query_id": "q1",
                "query_text": query_text,
                "kind": "gold",
            }
        ],
    }
    path = context.workdir / filename
    path.write_text(json.dumps(dataset, indent=2) + "\n", encoding="utf-8")
    context.dataset_path = path


@when('I create an evaluation dataset at "{filename}" with schema version {schema_version:d}')
def step_create_dataset_with_schema_version(context, filename: str, schema_version: int) -> None:
    dataset = {
        "schema_version": schema_version,
        "name": "schema-version-dataset",
        "description": "Dataset with custom schema version",
        "queries": [
            {
                "query_id": "q1",
                "query_text": "alpha",
                "expected_item_id": "placeholder",
                "kind": "gold",
            }
        ],
    }
    path = context.workdir / filename
    path.write_text(json.dumps(dataset, indent=2) + "\n", encoding="utf-8")
    context.dataset_path = path


@when('I attempt to evaluate the latest snapshot with dataset "{filename}"')
def step_eval_latest_run_invalid_dataset(context, filename: str) -> None:
    assert context.last_snapshot_id
    path = context.workdir / filename
    args = [
        "--corpus",
        str(_corpus_path(context, "corpus")),
        "eval",
        "--snapshot",
        context.last_snapshot_id,
        "--dataset",
        str(path),
        "--max-total-items",
        "3",
        "--maximum-total-characters",
        "2000",
        "--max-items-per-source",
        "5",
    ]
    run_biblicus(context, args)


@when('I download a Wikipedia corpus into "{corpus_name}"')
def step_download_wikipedia_corpus(context, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    result = subprocess.run(
        [sys.executable, "scripts/download_wikipedia.py", "--corpus", str(corpus), "--limit", "5"],
        cwd=context.repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stderr


@when('I download a Portable Document Format corpus into "{corpus_name}"')
def step_download_pdf_corpus(context, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    result = subprocess.run(
        [sys.executable, "scripts/download_pdf_samples.py", "--corpus", str(corpus), "--force"],
        cwd=context.repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stderr


@when('I download a mixed corpus into "{corpus_name}"')
def step_download_mixed_corpus(context, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    result = subprocess.run(
        [sys.executable, "scripts/download_mixed_samples.py", "--corpus", str(corpus), "--force"],
        cwd=context.repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stderr


@when('I download an audio corpus into "{corpus_name}"')
def step_download_audio_corpus(context, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    result = subprocess.run(
        [sys.executable, "scripts/download_audio_samples.py", "--corpus", str(corpus), "--force"],
        cwd=context.repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stderr


@when('I download an image corpus into "{corpus_name}"')
def step_download_image_corpus(context, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    result = subprocess.run(
        [sys.executable, "scripts/download_image_samples.py", "--corpus", str(corpus), "--force"],
        cwd=context.repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stderr


@when('I download a layout OCR corpus into "{corpus_name}"')
def step_download_layout_ocr_corpus(context, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    result = subprocess.run(
        [
            "python3",
            "scripts/download_funsd_samples.py",
            "--corpus",
            str(corpus),
            "--count",
            "3",
            "--force",
        ],
        cwd=context.repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stderr


@then("the corpus contains at least {count:d} items")
def step_corpus_contains_items(context, count: int) -> None:
    catalog = Corpus.open(_corpus_path(context, "corpus")).load_catalog()
    assert len(catalog.items) >= count


@then('the corpus contains at least {count:d} item with media type "{media_type}"')
@then('the corpus contains at least {count:d} items with media type "{media_type}"')
def step_corpus_contains_items_with_media_type(context, count: int, media_type: str) -> None:
    catalog = Corpus.open(_corpus_path(context, "corpus")).load_catalog()
    matching = [item for item in catalog.items.values() if item.media_type == media_type]
    assert len(matching) >= count


@then('the corpus contains at least {count:d} item tagged "{tag}"')
@then('the corpus contains at least {count:d} items tagged "{tag}"')
def step_corpus_contains_items_with_tag(context, count: int, tag: str) -> None:
    catalog = Corpus.open(_corpus_path(context, "corpus")).load_catalog()
    matching = [item for item in catalog.items.values() if tag in item.tags]
    assert len(matching) >= count



"""
Programmatic Wikipedia demo for Biblicus retrieval and context packs.
"""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Dict, Iterable, Tuple
from urllib.parse import quote
from urllib.request import Request, urlopen

from biblicus.context import (
    ContextPackPolicy,
    TokenBudget,
    build_context_pack,
    fit_context_pack_to_token_budget,
)
from biblicus.corpus import Corpus
from biblicus.extraction import build_extraction_snapshot
from biblicus.models import QueryBudget
from biblicus.retrievers import get_retriever

DEFAULT_TITLES = [
    "Retrieval-augmented generation",
    "Information retrieval",
    "Vector database",
    "Knowledge base",
    "Semantic search",
    "Search engine",
]


def fetch_wikipedia_summary(title: str) -> Tuple[str, str, str]:
    """
    Fetch a Wikipedia summary for a page title.

    :param title: Wikipedia page title.
    :type title: str
    :return: Tuple of extract text, page uniform resource locator, and resolved title.
    :rtype: tuple[str, str, str]
    """
    url = f"https://en.wikipedia.org/api/rest_v1/page/summary/{quote(title)}"
    request = Request(url, headers={"User-Agent": "biblicus/0"})
    with urlopen(request, timeout=30) as response:
        payload = json.loads(response.read().decode("utf-8"))
    extract = payload.get("extract") or ""
    resolved_title = payload.get("title") or title
    content_urls = payload.get("content_urls") or {}
    desktop_urls = content_urls.get("desktop") or {}
    page_url = desktop_urls.get("page") or f"https://en.wikipedia.org/wiki/{quote(resolved_title)}"
    return extract, page_url, resolved_title


def prepare_corpus(path: Path, *, force: bool) -> Corpus:
    """
    Initialize or open a corpus for the demo.

    :param path: Corpus path.
    :type path: Path
    :param force: Whether to purge existing corpus content.
    :type force: bool
    :return: Corpus instance.
    :rtype: Corpus
    :raises ValueError: If the target path is non-empty without force.
    """
    if (path / ".biblicus" / "config.json").is_file():
        corpus = Corpus.open(path)
        if force:
            corpus.purge(confirm=corpus.name)
        return corpus
    if path.exists() and any(path.iterdir()) and not force:
        raise ValueError("Target corpus directory is not empty. Use --force to initialize anyway.")
    return Corpus.init(path, force=True)


def ingest_wikipedia_summaries(corpus: Corpus, titles: Iterable[str]) -> Dict[str, int]:
    """
    Ingest Wikipedia summaries as notes in a corpus.

    :param corpus: Corpus to receive notes.
    :type corpus: Corpus
    :param titles: Wikipedia page titles.
    :type titles: Iterable[str]
    :return: Ingestion statistics.
    :rtype: dict[str, int]
    """
    ingested = 0
    skipped = 0
    for title in titles:
        extract, page_url, resolved_title = fetch_wikipedia_summary(title)
        if not extract.strip():
            skipped += 1
            continue
        corpus.ingest_note(
            extract,
            title=resolved_title,
            tags=["wikipedia", "rag-demo"],
            source_uri=page_url,
        )
        ingested += 1
    corpus.reindex()
    return {"ingested": ingested, "skipped": skipped}


def run_demo(*, corpus_path: Path, limit: int, force: bool, query: str) -> Dict[str, object]:
    """
    Run the end-to-end Wikipedia demo.

    :param corpus_path: Corpus path to create or reuse.
    :type corpus_path: Path
    :param limit: Number of titles to ingest.
    :type limit: int
    :param force: Whether to purge existing corpus content.
    :type force: bool
    :param query: Retrieval query text.
    :type query: str
    :return: Summary output including query results and context pack text.
    :rtype: dict[str, object]
    """
    corpus = prepare_corpus(corpus_path, force=force)
    titles = DEFAULT_TITLES[:limit]
    ingest_stats = ingest_wikipedia_summaries(corpus, titles)

    extraction_manifest = build_extraction_snapshot(
        corpus,
        extractor_id="pipeline",
        configuration_name="Wikipedia demo extraction",
        configuration={"stages": [{"extractor_id": "pass-through-text", "config": {}}]},
    )

    backend = get_retriever("sqlite-full-text-search")
    snapshot = backend.build_snapshot(
        corpus,
        configuration_name="Wikipedia demo retrieval",
        configuration={
            "extraction_snapshot": f"pipeline:{extraction_manifest.snapshot_id}",
            "chunk_size": 220,
            "chunk_overlap": 40,
            "snippet_characters": 160,
        },
    )
    budget = QueryBudget(max_total_items=5, maximum_total_characters=1600, max_items_per_source=1)
    retrieval_result = backend.query(corpus, snapshot=snapshot, query_text=query, budget=budget)

    policy = ContextPackPolicy(join_with="\n\n")
    context_pack = build_context_pack(retrieval_result, policy=policy)
    token_budget = TokenBudget(max_tokens=400)
    fitted_context_pack = fit_context_pack_to_token_budget(
        context_pack,
        policy=policy,
        token_budget=token_budget,
    )

    return {
        "ingest_stats": ingest_stats,
        "extraction_snapshot_id": extraction_manifest.snapshot_id,
        "retrieval_snapshot_id": snapshot.snapshot_id,
        "query_text": query,
        "evidence": [evidence.model_dump() for evidence in retrieval_result.evidence],
        "context_pack_text": fitted_context_pack.text,
    }


def build_parser() -> argparse.ArgumentParser:
    """
    Build the command-line interface argument parser.

    :return: Argument parser.
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(description="Run the Wikipedia retrieval demo.")
    parser.add_argument("--corpus", required=True, help="Corpus path to initialize or reuse.")
    parser.add_argument(
        "--limit", type=int, default=5, help="Number of Wikipedia pages to download."
    )
    parser.add_argument("--force", action="store_true", help="Purge existing corpus content.")
    parser.add_argument(
        "--query",
        default="retrieval augmented generation",
        help="Query text to run against the corpus.",
    )
    return parser


def main() -> int:
    """
    Entry point for the Wikipedia retrieval demo.

    :return: Exit code.
    :rtype: int
    """
    parser = build_parser()
    args = parser.parse_args()
    output = run_demo(
        corpus_path=Path(args.corpus).resolve(),
        limit=args.limit,
        force=args.force,
        query=args.query,
    )
    print(json.dumps(output, indent=2))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""
Hybrid retriever combining lexical and vector results.
"""

from __future__ import annotations

from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, model_validator

from ..corpus import Corpus
from ..models import Evidence, QueryBudget, RetrievalResult, RetrievalSnapshot
from ..retrieval import apply_budget, create_configuration_manifest, create_snapshot_manifest
from ..time import utc_now_iso


class HybridConfiguration(BaseModel):
    """
    Configuration for hybrid retrieval fusion.

    :ivar lexical_retriever: Retriever identifier for lexical retrieval.
    :vartype lexical_retriever: str
    :ivar embedding_retriever: Retriever identifier for embedding retrieval.
    :vartype embedding_retriever: str
    :ivar lexical_weight: Weight for lexical scores.
    :vartype lexical_weight: float
    :ivar embedding_weight: Weight for embedding scores.
    :vartype embedding_weight: float
    :ivar lexical_configuration: Optional lexical retriever configuration.
    :vartype lexical_configuration: dict[str, object]
    :ivar embedding_configuration: Optional embedding retriever configuration.
    :vartype embedding_configuration: dict[str, object]
    """

    model_config = ConfigDict(extra="forbid")

    lexical_retriever: str = Field(default="sqlite-full-text-search", min_length=1)
    embedding_retriever: str = Field(default="tf-vector", min_length=1)
    lexical_weight: float = Field(default=0.5, ge=0, le=1)
    embedding_weight: float = Field(default=0.5, ge=0, le=1)
    lexical_configuration: Dict[str, object] = Field(default_factory=dict)
    embedding_configuration: Dict[str, object] = Field(default_factory=dict)

    @model_validator(mode="after")
    def _validate_weights(self) -> "HybridConfiguration":
        if abs((self.lexical_weight + self.embedding_weight) - 1.0) > 1e-6:
            raise ValueError("weights must sum to 1")
        return self


class HybridRetriever:
    """
    Hybrid retriever that fuses lexical and embedding retrieval.

    :ivar retriever_id: Retriever identifier.
    :vartype retriever_id: str
    """

    retriever_id = "hybrid"

    def build_snapshot(
        self, corpus: Corpus, *, configuration_name: str, configuration: Dict[str, object]
    ) -> RetrievalSnapshot:
        """
        Build or register a hybrid retrieval snapshot.

        :param corpus: Corpus to build against.
        :type corpus: Corpus
        :param configuration_name: Human-readable configuration name.
        :type configuration_name: str
        :param configuration: Retriever-specific configuration values.
        :type configuration: dict[str, object]
        :return: Snapshot manifest describing the build.
        :rtype: RetrievalSnapshot
        """
        parsed_config = HybridConfiguration.model_validate(configuration)
        _ensure_retriever_supported(parsed_config)
        lexical_retriever = _resolve_retriever(parsed_config.lexical_retriever)
        embedding_retriever = _resolve_retriever(parsed_config.embedding_retriever)
        lexical_snapshot = lexical_retriever.build_snapshot(
            corpus,
            configuration_name=f"{configuration_name}-lexical",
            configuration=parsed_config.lexical_configuration,
        )
        embedding_snapshot = embedding_retriever.build_snapshot(
            corpus,
            configuration_name=f"{configuration_name}-embedding",
            configuration=parsed_config.embedding_configuration,
        )
        configuration_manifest = create_configuration_manifest(
            retriever_id=self.retriever_id,
            name=configuration_name,
            configuration=parsed_config.model_dump(),
        )
        stats = {
            "lexical_snapshot_id": lexical_snapshot.snapshot_id,
            "embedding_snapshot_id": embedding_snapshot.snapshot_id,
        }
        snapshot = create_snapshot_manifest(
            corpus,
            configuration=configuration_manifest,
            stats=stats,
            snapshot_artifacts=[],
        )
        corpus.write_snapshot(snapshot)
        return snapshot

    def query(
        self,
        corpus: Corpus,
        *,
        snapshot: RetrievalSnapshot,
        query_text: str,
        budget: QueryBudget,
    ) -> RetrievalResult:
        """
        Query using both lexical and embedding retrievers and fuse scores.

        :param corpus: Corpus associated with the snapshot.
        :type corpus: Corpus
        :param snapshot: Snapshot manifest to use for querying.
        :type snapshot: RetrievalSnapshot
        :param query_text: Query text to execute.
        :type query_text: str
        :param budget: Evidence selection budget.
        :type budget: QueryBudget
        :return: Retrieval results containing evidence.
        :rtype: RetrievalResult
        """
        configuration = HybridConfiguration.model_validate(snapshot.configuration.configuration)
        _ensure_retriever_supported(configuration)
        lexical_retriever = _resolve_retriever(configuration.lexical_retriever)
        embedding_retriever = _resolve_retriever(configuration.embedding_retriever)
        lexical_snapshot_id = snapshot.stats.get("lexical_snapshot_id")
        embedding_snapshot_id = snapshot.stats.get("embedding_snapshot_id")
        if not lexical_snapshot_id or not embedding_snapshot_id:
            raise ValueError("Hybrid snapshot missing lexical or embedding snapshot identifiers")
        lexical_snapshot = corpus.load_snapshot(str(lexical_snapshot_id))
        embedding_snapshot = corpus.load_snapshot(str(embedding_snapshot_id))
        component_budget = _expand_component_budget(budget)
        lexical_result = lexical_retriever.query(
            corpus, snapshot=lexical_snapshot, query_text=query_text, budget=component_budget
        )
        embedding_result = embedding_retriever.query(
            corpus, snapshot=embedding_snapshot, query_text=query_text, budget=component_budget
        )
        candidates = _fuse_evidence(
            lexical_result.evidence,
            embedding_result.evidence,
            lexical_weight=configuration.lexical_weight,
            embedding_weight=configuration.embedding_weight,
        )
        sorted_candidates = sorted(
            candidates,
            key=lambda evidence_item: (-evidence_item.score, evidence_item.item_id),
        )
        ranked = [
            evidence_item.model_copy(
                update={
                    "rank": index,
                    "configuration_id": snapshot.configuration.configuration_id,
                    "snapshot_id": snapshot.snapshot_id,
                }
            )
            for index, evidence_item in enumerate(sorted_candidates, start=1)
        ]
        evidence = apply_budget(ranked, budget)
        stats = {
            "candidates": len(sorted_candidates),
            "returned": len(evidence),
            "fusion_weights": {
                "lexical": configuration.lexical_weight,
                "embedding": configuration.embedding_weight,
            },
        }
        return RetrievalResult(
            query_text=query_text,
            budget=budget,
            snapshot_id=snapshot.snapshot_id,
            configuration_id=snapshot.configuration.configuration_id,
            retriever_id=snapshot.configuration.retriever_id,
            generated_at=utc_now_iso(),
            evidence=evidence,
            stats=stats,
        )


def _ensure_retriever_supported(configuration: HybridConfiguration) -> None:
    """
    Validate that hybrid retrievers do not reference the hybrid retriever itself.

    :param configuration: Parsed hybrid configuration.
    :type configuration: HybridConfiguration
    :return: None.
    :rtype: None
    :raises ValueError: If hybrid is used as a component retriever.
    """
    if configuration.lexical_retriever == HybridRetriever.retriever_id:
        raise ValueError("Hybrid retriever cannot use itself as the lexical retriever")
    if configuration.embedding_retriever == HybridRetriever.retriever_id:
        raise ValueError("Hybrid retriever cannot use itself as the embedding retriever")


def _resolve_retriever(retriever_id: str):
    """
    Resolve a retriever by identifier.

    :param retriever_id: Retriever identifier.
    :type retriever_id: str
    :return: Retriever instance.
    :rtype: object
    """
    from biblicus.retrievers import get_retriever  # Delayed import to avoid circular import

    return get_retriever(retriever_id)


def _expand_component_budget(budget: QueryBudget, *, multiplier: int = 5) -> QueryBudget:
    """
    Expand a final budget to collect more candidates for fusion.

    :param budget: Final evidence budget.
    :type budget: QueryBudget
    :param multiplier: Candidate expansion multiplier.
    :type multiplier: int
    :return: Expanded budget for component retrievers.
    :rtype: QueryBudget
    """
    maximum_total_characters = budget.maximum_total_characters
    expanded_characters = (
        maximum_total_characters * multiplier if maximum_total_characters is not None else None
    )
    expanded_max_items_per_source = (
        budget.max_items_per_source * multiplier
        if budget.max_items_per_source is not None
        else None
    )
    requested_items = budget.max_total_items + budget.offset
    return QueryBudget(
        max_total_items=requested_items * multiplier,
        offset=0,
        maximum_total_characters=expanded_characters,
        max_items_per_source=expanded_max_items_per_source,
    )


def _fuse_evidence(
    lexical: List[Evidence],
    embedding: List[Evidence],
    *,
    lexical_weight: float,
    embedding_weight: float,
) -> List[Evidence]:
    """
    Fuse lexical and embedding evidence lists into hybrid candidates.

    :param lexical: Lexical evidence list.
    :type lexical: list[Evidence]
    :param embedding: Embedding evidence list.
    :type embedding: list[Evidence]
    :param lexical_weight: Lexical score weight.
    :type lexical_weight: float
    :param embedding_weight: Embedding score weight.
    :type embedding_weight: float
    :return: Hybrid evidence list.
    :rtype: list[Evidence]
    """
    merged: Dict[str, Dict[str, Optional[Evidence]]] = {}
    for evidence_item in lexical:
        merged.setdefault(evidence_item.item_id, {})["lexical"] = evidence_item
    for evidence_item in embedding:
        merged.setdefault(evidence_item.item_id, {})["embedding"] = evidence_item

    candidates: List[Evidence] = []
    for item_id, sources in merged.items():
        lexical_evidence = sources.get("lexical")
        embedding_evidence = sources.get("embedding")
        lexical_score = lexical_evidence.score if lexical_evidence else 0.0
        embedding_score = embedding_evidence.score if embedding_evidence else 0.0
        combined_score = (lexical_score * lexical_weight) + (embedding_score * embedding_weight)
        base_evidence = lexical_evidence or embedding_evidence
        candidates.append(
            Evidence(
                item_id=item_id,
                source_uri=base_evidence.source_uri,
                media_type=base_evidence.media_type,
                score=combined_score,
                rank=1,
                text=base_evidence.text,
                content_ref=base_evidence.content_ref,
                span_start=base_evidence.span_start,
                span_end=base_evidence.span_end,
                stage="hybrid",
                stage_scores={"lexical": lexical_score, "embedding": embedding_score},
                configuration_id="",
                snapshot_id="",
                metadata=base_evidence.metadata,
                hash=base_evidence.hash,
            )
        )
    return candidates



"""
Context pack building for Biblicus.

A context pack is the text that your application sends to a large language model.
Biblicus produces a context pack from structured retrieval results so that evidence remains a
stable contract while context formatting remains an explicit policy surface.
"""

from __future__ import annotations

from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from .models import Evidence, RetrievalResult


class ContextPackPolicy(BaseModel):
    """
    Policy that controls how evidence becomes context pack text.

    :ivar join_with: Separator inserted between evidence text blocks.
    :vartype join_with: str
    :ivar ordering: Evidence ordering policy (rank, score, or source).
    :vartype ordering: str
    :ivar include_metadata: Whether to include evidence metadata lines in each block.
    :vartype include_metadata: bool
    :ivar metadata_fields: Optional evidence metadata fields to include.
    :vartype metadata_fields: list[str] or None
    """

    model_config = ConfigDict(extra="forbid")

    join_with: str = Field(default="\n\n")
    ordering: str = Field(default="rank", min_length=1)
    include_metadata: bool = Field(default=False)
    metadata_fields: Optional[List[str]] = None


class ContextPack(BaseModel):
    """
    Context pack derived from retrieval evidence.

    :ivar text: Context pack text suitable for inclusion in a model call.
    :vartype text: str
    :ivar evidence_count: Number of evidence blocks included in the context pack.
    :vartype evidence_count: int
    :ivar blocks: Structured blocks that produced the context pack.
    :vartype blocks: list[ContextPackBlock]
    """

    model_config = ConfigDict(extra="forbid")

    text: str
    evidence_count: int = Field(ge=0)
    blocks: List["ContextPackBlock"] = Field(default_factory=list)


class ContextPackBlock(BaseModel):
    """
    A single context pack block derived from one evidence item.

    :ivar evidence_item_id: Item identifier that produced this block.
    :vartype evidence_item_id: str
    :ivar text: Text included in this block.
    :vartype text: str
    :ivar metadata: Optional metadata included with the block.
    :vartype metadata: dict[str, object] or None
    """

    model_config = ConfigDict(extra="forbid")

    evidence_item_id: str = Field(min_length=1)
    text: str = Field(min_length=1)
    metadata: Optional[Dict[str, object]] = None


class TokenCounter(BaseModel):
    """
    Token counter configuration for token budget fitting.

    This is a lightweight model wrapper so token fitting remains explicit and testable even when
    the underlying tokenizer is provided by an optional dependency.

    :ivar tokenizer_id: Tokenizer identifier (for example, naive-whitespace).
    :vartype tokenizer_id: str
    """

    model_config = ConfigDict(extra="forbid")

    tokenizer_id: str = Field(default="naive-whitespace", min_length=1)


class TokenBudget(BaseModel):
    """
    Token budget for a context pack.

    :ivar max_tokens: Maximum tokens permitted for the final context pack text.
    :vartype max_tokens: int
    """

    model_config = ConfigDict(extra="forbid")

    max_tokens: int = Field(ge=1)


class CharacterBudget(BaseModel):
    """
    Character budget for a context pack.

    :ivar max_characters: Maximum characters permitted for the final context pack text.
    :vartype max_characters: int
    """

    model_config = ConfigDict(extra="forbid")

    max_characters: int = Field(ge=1)


def build_context_pack(result: RetrievalResult, *, policy: ContextPackPolicy) -> ContextPack:
    """
    Build a context pack from a retrieval result using an explicit policy.

    :param result: Retrieval result containing ranked evidence.
    :type result: RetrievalResult
    :param policy: Policy controlling how evidence text is joined.
    :type policy: ContextPackPolicy
    :return: Context pack containing concatenated evidence text.
    :rtype: ContextPack
    """
    selected_blocks: List[ContextPackBlock] = []
    for evidence in _order_evidence(result.evidence, policy=policy):
        if not isinstance(evidence.text, str):
            continue
        trimmed_text = evidence.text.strip()
        if not trimmed_text:
            continue
        metadata = (
            _metadata_for_evidence(evidence, policy=policy) if policy.include_metadata else None
        )
        block_text = _format_block_text(trimmed_text, metadata=metadata)
        selected_blocks.append(
            ContextPackBlock(
                evidence_item_id=evidence.item_id,
                text=block_text,
                metadata=metadata,
            )
        )

    return ContextPack(
        text=policy.join_with.join([block.text for block in selected_blocks]),
        evidence_count=len(selected_blocks),
        blocks=selected_blocks,
    )


def count_tokens(text: str, *, tokenizer_id: str) -> int:
    """
    Count tokens in a text using a tokenizer identifier.

    The default tokenizer is naive-whitespace, which counts whitespace-separated tokens.

    :param text: Text payload to count.
    :type text: str
    :param tokenizer_id: Tokenizer identifier.
    :type tokenizer_id: str
    :return: Token count.
    :rtype: int
    :raises KeyError: If the tokenizer identifier is unknown.
    """
    tokenizers = {
        "naive-whitespace": lambda value: len([token for token in value.split() if token]),
    }
    tokenizer = tokenizers[tokenizer_id]
    return int(tokenizer(text))


def fit_context_pack_to_token_budget(
    context_pack: ContextPack,
    *,
    policy: ContextPackPolicy,
    token_budget: TokenBudget,
    token_counter: Optional[TokenCounter] = None,
) -> ContextPack:
    """
    Fit a context pack to a token budget by dropping trailing blocks.

    This function is deterministic. It never rewrites block text. It only removes blocks from the
    end of the block list until the token budget is met.

    :param context_pack: Context pack to fit.
    :type context_pack: ContextPack
    :param policy: Policy controlling how blocks are joined into text.
    :type policy: ContextPackPolicy
    :param token_budget: Token budget to enforce.
    :type token_budget: TokenBudget
    :param token_counter: Optional token counter configuration.
    :type token_counter: TokenCounter or None
    :return: Fitted context pack.
    :rtype: ContextPack
    """
    token_counter = token_counter or TokenCounter()
    remaining_blocks: List[ContextPackBlock] = list(context_pack.blocks)

    while remaining_blocks:
        candidate_text = policy.join_with.join([block.text for block in remaining_blocks])
        candidate_tokens = count_tokens(candidate_text, tokenizer_id=token_counter.tokenizer_id)
        if candidate_tokens <= token_budget.max_tokens:
            return ContextPack(
                text=candidate_text,
                evidence_count=len(remaining_blocks),
                blocks=remaining_blocks,
            )
        remaining_blocks = remaining_blocks[:-1]

    return ContextPack(text="", evidence_count=0, blocks=[])


def fit_context_pack_to_character_budget(
    context_pack: ContextPack,
    *,
    policy: ContextPackPolicy,
    character_budget: CharacterBudget,
) -> ContextPack:
    """
    Fit a context pack to a character budget by dropping trailing blocks.

    :param context_pack: Context pack to fit.
    :type context_pack: ContextPack
    :param policy: Policy controlling how blocks are joined into text.
    :type policy: ContextPackPolicy
    :param character_budget: Character budget to enforce.
    :type character_budget: CharacterBudget
    :return: Fitted context pack.
    :rtype: ContextPack
    """
    remaining_blocks: List[ContextPackBlock] = list(context_pack.blocks)
    max_characters = character_budget.max_characters

    while remaining_blocks:
        candidate_text = policy.join_with.join([block.text for block in remaining_blocks])
        if len(candidate_text) <= max_characters:
            return ContextPack(
                text=candidate_text,
                evidence_count=len(remaining_blocks),
                blocks=remaining_blocks,
            )
        remaining_blocks = remaining_blocks[:-1]

    return ContextPack(text="", evidence_count=0, blocks=[])


def _order_evidence(
    evidence: List[Evidence],
    *,
    policy: ContextPackPolicy,
) -> List[Evidence]:
    """
    Order evidence items according to the context pack policy.

    :param evidence: Evidence list to order.
    :type evidence: list[Evidence]
    :param policy: Context pack policy.
    :type policy: ContextPackPolicy
    :return: Ordered evidence list.
    :rtype: list[Evidence]
    """
    if policy.ordering == "rank":
        return sorted(evidence, key=lambda item: (item.rank, item.item_id))
    if policy.ordering == "score":
        return sorted(evidence, key=lambda item: (-item.score, item.item_id))
    if policy.ordering == "source":
        return sorted(
            evidence,
            key=lambda item: (
                item.source_uri or item.item_id,
                -item.score,
                item.item_id,
            ),
        )
    raise ValueError(f"Unknown context pack ordering: {policy.ordering}")


def _metadata_for_evidence(
    evidence: Evidence,
    *,
    policy: ContextPackPolicy,
) -> Dict[str, object]:
    """
    Build metadata for a context pack block.

    :param evidence: Evidence item to describe.
    :type evidence: Evidence
    :return: Metadata mapping.
    :rtype: dict[str, object]
    """
    metadata = {
        "item_id": evidence.item_id,
        "source_uri": evidence.source_uri or "none",
        "score": evidence.score,
        "stage": evidence.stage,
    }
    extra = evidence.metadata or {}
    if policy.metadata_fields is not None:
        extra = {key: extra.get(key) for key in policy.metadata_fields if key in extra}
    for key, value in extra.items():
        if key not in metadata:
            metadata[key] = value
    return metadata


def _format_block_text(text: str, *, metadata: Optional[Dict[str, object]]) -> str:
    """
    Format a context pack block text with optional metadata.

    :param text: Evidence text.
    :type text: str
    :param metadata: Optional metadata mapping.
    :type metadata: dict[str, object] or None
    :return: Formatted block text.
    :rtype: str
    """
    if not metadata:
        return text
    ordered_keys = ["item_id", "source_uri", "score", "stage"]
    metadata_lines = [f"{key}: {metadata[key]}" for key in ordered_keys if key in metadata]
    for key in sorted(metadata.keys()):
        if key in ordered_keys:
            continue
        metadata_lines.append(f"{key}: {metadata[key]}")
    metadata_text = "\n".join(metadata_lines)
    return f"{metadata_text}\n{text}"



"""
Embedding helpers for Reinforcement Memory.

Defines the :data:`EmbedFn` callable type and provides factory functions
that return it:

- :func:`hash_embedder` — deterministic hash-based embeddings, no external
  services.  Useful for tests and demos.
- :func:`sentence_transformer_embedder` — sentence-transformers with optional
  S3 or local caching.  Requires ``sentence-transformers``.
- :func:`dspy_embedder` — Biblicus's DSPy-based embedding backend.  Requires
  the ``dspy`` optional dependency group.

Also provides :class:`EmbeddingCacheProtocol`, :class:`S3EmbeddingCache`, and
:class:`LocalEmbeddingCache` for the caching layer used by
:func:`sentence_transformer_embedder`.
"""

from __future__ import annotations

import hashlib
import json
import logging
import os
import re
from typing import Callable, List, Optional, Protocol, runtime_checkable

import numpy as np

logger = logging.getLogger(__name__)

# Type alias for an embedding function accepted by :class:`ReinforcementMemory`.
EmbedFn = Callable[[List[str]], np.ndarray]

_DEFAULT_MODEL_ID = "all-MiniLM-L6-v2"
_DEFAULT_PREPROCESSING_VERSION = "1"


# ---------------------------------------------------------------------------
# Embedding cache protocol + implementations
# ---------------------------------------------------------------------------


@runtime_checkable
class EmbeddingCacheProtocol(Protocol):
    """Protocol for an embedding cache backend."""

    def get(self, model_id: str, key: str) -> Optional[np.ndarray]:
        """Return cached embedding or None on miss."""
        ...

    def put(self, model_id: str, key: str, embedding: np.ndarray) -> None:
        """Store an embedding (non-fatal on failure)."""
        ...


def _normalize_text(text: str) -> str:
    """Collapse whitespace and lowercase for deterministic cache keys."""
    if not text or not isinstance(text, str):
        return ""
    return re.sub(r"\s+", " ", text.strip()).lower()


def _cache_key(text: str, model_id: str, version: str) -> str:
    """Return SHA-256 cache key for a text + model + version triple."""
    payload = f"{_normalize_text(text)}|{model_id}|{version}"
    return hashlib.sha256(payload.encode("utf-8")).hexdigest()


def _s3_path(model_id: str, key: str) -> str:
    prefix = key[:2] if len(key) >= 2 else "00"
    return f"embeddings/{model_id}/{prefix}/{key}.json"


class S3EmbeddingCache:
    """
    S3-backed embedding cache.

    Cache key: ``SHA-256(normalize(text)|model_id|version)``
    S3 path: ``embeddings/{model_id}/{key_prefix}/{key}.json``

    :param bucket_name: S3 bucket for cached embeddings.
    :param s3_client: Optional pre-built boto3 S3 client.
    """

    def __init__(
        self,
        bucket_name: str,
        s3_client=None,
    ) -> None:
        """Initialise the S3 cache."""
        self.bucket_name = bucket_name
        if s3_client is not None:
            self._s3 = s3_client
        else:
            import boto3

            self._s3 = boto3.client("s3")

    def get(self, model_id: str, key: str) -> Optional[np.ndarray]:
        """Return cached embedding or None on miss."""
        try:
            path = _s3_path(model_id, key)
            resp = self._s3.get_object(Bucket=self.bucket_name, Key=path)
            data = json.loads(resp["Body"].read().decode("utf-8"))
            vec = data.get("embedding")
            return np.array(vec, dtype=np.float32) if vec is not None else None
        except Exception:
            return None

    def put(self, model_id: str, key: str, embedding: np.ndarray) -> None:
        """Write embedding to S3 (non-fatal on failure)."""
        try:
            path = _s3_path(model_id, key)
            body = json.dumps({"embedding": embedding.tolist()}).encode("utf-8")
            self._s3.put_object(
                Bucket=self.bucket_name,
                Key=path,
                Body=body,
                ContentType="application/json",
            )
        except Exception as exc:
            logger.warning("S3EmbeddingCache put failed (non-fatal): %s", exc)


class LocalEmbeddingCache:
    """
    Filesystem-backed embedding cache.

    Cache files are stored as JSON under ``cache_dir/{model_id}/{prefix}/{key}.json``.

    :param cache_dir: Root directory for cached embeddings.
    """

    def __init__(self, cache_dir: str) -> None:
        """Initialise the local cache."""
        self._root = cache_dir

    def get(self, model_id: str, key: str) -> Optional[np.ndarray]:
        """Return cached embedding or None on miss."""
        path = self._path(model_id, key)
        try:
            with open(path, encoding="utf-8") as fh:
                data = json.load(fh)
            vec = data.get("embedding")
            return np.array(vec, dtype=np.float32) if vec is not None else None
        except (FileNotFoundError, json.JSONDecodeError, Exception):
            return None

    def put(self, model_id: str, key: str, embedding: np.ndarray) -> None:
        """Write embedding to disk (non-fatal on failure)."""
        path = self._path(model_id, key)
        try:
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "w", encoding="utf-8") as fh:
                json.dump({"embedding": embedding.tolist()}, fh)
        except Exception as exc:
            logger.warning("LocalEmbeddingCache put failed (non-fatal): %s", exc)

    def _path(self, model_id: str, key: str) -> str:
        prefix = key[:2] if len(key) >= 2 else "00"
        safe_model = model_id.replace("/", "_")
        return os.path.join(self._root, safe_model, prefix, f"{key}.json")


# ---------------------------------------------------------------------------
# Embedding factory helpers
# ---------------------------------------------------------------------------


def hash_embedder(dimensions: int = 384) -> EmbedFn:
    """
    Return an :data:`EmbedFn` that produces deterministic hash-based embeddings.

    No external services or models required.  Suitable for tests and demos.

    :param dimensions: Output vector dimension.
    :return: Callable that maps a list of strings to an ndarray of shape
        ``(n, dimensions)``.
    """

    def _embed(texts: List[str]) -> np.ndarray:
        result = []
        for text in texts:
            seed = int(hashlib.md5(text.encode("utf-8")).hexdigest(), 16) % (2**32)
            rng = np.random.default_rng(seed)
            vec = rng.standard_normal(dimensions).astype(np.float32)
            norm = np.linalg.norm(vec)
            result.append(vec / norm if norm > 0 else vec)
        return np.array(result, dtype=np.float32)

    return _embed


def sentence_transformer_embedder(
    model_id: str = _DEFAULT_MODEL_ID,
    cache: Optional[EmbeddingCacheProtocol] = None,
    preprocessing_version: str = _DEFAULT_PREPROCESSING_VERSION,
) -> EmbedFn:
    """
    Return an :data:`EmbedFn` backed by sentence-transformers.

    Results are cached via ``cache`` when provided, avoiding re-embedding
    identical texts across runs.

    Requires ``sentence-transformers`` (``pip install biblicus[sentence-transformers]``).

    :param model_id: Sentence-transformers model name (default
        ``"all-MiniLM-L6-v2"``).
    :param cache: Optional :class:`EmbeddingCacheProtocol` implementation.
        Use :class:`S3EmbeddingCache` for production or
        :class:`LocalEmbeddingCache` for local development.
    :param preprocessing_version: Version string included in the cache key.
        Increment this to invalidate the cache after preprocessing changes.
    :return: Callable that maps a list of strings to an ndarray of shape
        ``(n, dim)``.
    """
    _model_holder: list = []  # mutable container for lazy singleton

    def _load_model():
        if not _model_holder:
            try:
                from sentence_transformers import SentenceTransformer
            except ImportError as exc:
                raise ImportError(
                    "sentence_transformer_embedder requires sentence-transformers. "
                    'Install it with: pip install "biblicus[sentence-transformers]"'
                ) from exc
            _model_holder.append(SentenceTransformer(model_id))
        return _model_holder[0]

    def _embed(texts: List[str]) -> np.ndarray:
        keys = [_cache_key(t, model_id, preprocessing_version) for t in texts]
        result: list = [None] * len(texts)
        miss_indices: list = []
        miss_texts: list = []

        if cache is not None:
            for i, (text, key) in enumerate(zip(texts, keys)):
                cached = cache.get(model_id, key)
                if cached is not None:
                    result[i] = cached
                else:
                    miss_indices.append(i)
                    miss_texts.append(text)
        else:
            miss_indices = list(range(len(texts)))
            miss_texts = list(texts)

        if miss_texts:
            model = _load_model()
            embeddings = model.encode(miss_texts, convert_to_numpy=True)
            for idx, emb in zip(miss_indices, embeddings):
                arr = np.asarray(emb, dtype=np.float32)
                result[idx] = arr
                if cache is not None:
                    cache.put(model_id, keys[idx], arr)

        return np.array(result, dtype=np.float32)

    return _embed


def dspy_embedder(client) -> EmbedFn:
    """
    Return an :data:`EmbedFn` backed by Biblicus's DSPy embedding backend.

    :param client: :class:`~biblicus.ai.models.EmbeddingsClientConfig` instance.
    :return: Callable that maps a list of strings to an ndarray.
    """
    from biblicus.ai.embeddings import generate_embeddings_batch

    def _embed(texts: List[str]) -> np.ndarray:
        vectors = generate_embeddings_batch(client=client, texts=texts)
        return np.array(vectors, dtype=np.float32)

    return _embed



from __future__ import annotations

import sys
import types
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from urllib.parse import unquote

from behave import given, then


@dataclass
class _FakeWhisperSegment:
    text: str


@dataclass
class _FakeWhisperInfo:
    language: str
    language_probability: float
    duration: float


@dataclass
class _FakeWhisperBehavior:
    segments: List[_FakeWhisperSegment]
    info: _FakeWhisperInfo


def _ensure_fake_whisper_behaviors(context) -> Dict[str, _FakeWhisperBehavior]:
    behaviors = getattr(context, "fake_whisper_behaviors", None)
    if behaviors is None:
        behaviors = {}
        context.fake_whisper_behaviors = behaviors
    return behaviors


def _install_fake_faster_whisper_module(context) -> None:
    already_installed = getattr(context, "_fake_faster_whisper_installed", False)
    if already_installed:
        return

    original_modules: Dict[str, object] = {}
    if "faster_whisper" in sys.modules:
        original_modules["faster_whisper"] = sys.modules["faster_whisper"]

    behaviors = _ensure_fake_whisper_behaviors(context)

    class WhisperModel:
        def __init__(self, model_size: str, device: str = "auto", compute_type: str = "int8") -> None:
            faster_whisper_module.last_model_size = model_size  # type: ignore[attr-defined]
            faster_whisper_module.last_device = device  # type: ignore[attr-defined]
            faster_whisper_module.last_compute_type = compute_type  # type: ignore[attr-defined]

        def transcribe(
            self,
            audio_path: str,
            language: Optional[str] = None,
            beam_size: int = 5
        ) -> tuple[List[Any], Any]:
            faster_whisper_module.last_audio_path = audio_path  # type: ignore[attr-defined]
            faster_whisper_module.last_language = language  # type: ignore[attr-defined]
            faster_whisper_module.last_beam_size = beam_size  # type: ignore[attr-defined]

            # Extract filename from path
            base_filename = audio_path.rsplit("/", 1)[-1]
            # Strip the UUID prefix (format: uuid--originalfile)
            filename = base_filename.split("--", 1)[-1] if "--" in base_filename else base_filename
            # URL decode if needed
            filename = unquote(filename)
            # Extract just the basename (in case it's a full file:// URI)
            filename = filename.rsplit("/", 1)[-1]

            # Find behavior for this filename
            behavior = behaviors.get(filename)
            if behavior is None:
                # Default behavior: empty transcript
                behavior = _FakeWhisperBehavior(
                    segments=[],
                    info=_FakeWhisperInfo(
                        language="en",
                        language_probability=0.99,
                        duration=0.0
                    )
                )

            return behavior.segments, behavior.info

    faster_whisper_module = types.ModuleType("faster_whisper")
    faster_whisper_module.WhisperModel = WhisperModel
    faster_whisper_module.last_model_size = None
    faster_whisper_module.last_device = None
    faster_whisper_module.last_compute_type = None
    faster_whisper_module.last_audio_path = None
    faster_whisper_module.last_language = None
    faster_whisper_module.last_beam_size = None

    sys.modules["faster_whisper"] = faster_whisper_module

    context._fake_faster_whisper_installed = True
    context._fake_faster_whisper_original_modules = original_modules


def _install_faster_whisper_unavailable_module(context) -> None:
    already_installed = getattr(context, "_fake_faster_whisper_unavailable_installed", False)
    if already_installed:
        return

    original_modules: Dict[str, object] = {}
    if "faster_whisper" in sys.modules:
        original_modules["faster_whisper"] = sys.modules["faster_whisper"]

    faster_whisper_module = types.ModuleType("faster_whisper")
    sys.modules["faster_whisper"] = faster_whisper_module

    context._fake_faster_whisper_unavailable_installed = True
    context._fake_faster_whisper_unavailable_original_modules = original_modules


@given("a fake faster-whisper library is available")
def step_fake_faster_whisper_available(context) -> None:
    _install_fake_faster_whisper_module(context)


@given(
    'a fake faster-whisper library is available that returns transcript "{transcript}" for filename "{filename}"'
)
def step_fake_faster_whisper_returns_transcript(context, transcript: str, filename: str) -> None:
    _install_fake_faster_whisper_module(context)
    behaviors = _ensure_fake_whisper_behaviors(context)

    # Split transcript into segments (by sentence)
    segments = [_FakeWhisperSegment(text=transcript)]

    info = _FakeWhisperInfo(
        language="en",
        language_probability=0.99,
        duration=5.0
    )

    behaviors[filename] = _FakeWhisperBehavior(segments=segments, info=info)


@given(
    'a fake faster-whisper library is available that returns transcript "{transcript}" with detected language "{language}" for file "{filename}"'
)
def step_fake_faster_whisper_returns_transcript_with_language(
    context, transcript: str, language: str, filename: str
) -> None:
    _install_fake_faster_whisper_module(context)
    behaviors = _ensure_fake_whisper_behaviors(context)

    segments = [_FakeWhisperSegment(text=transcript)]
    info = _FakeWhisperInfo(
        language=language,
        language_probability=0.95,
        duration=5.0
    )

    behaviors[filename] = _FakeWhisperBehavior(segments=segments, info=info)


@given("the faster-whisper dependency is unavailable")
def step_faster_whisper_dependency_unavailable(context) -> None:
    _install_faster_whisper_unavailable_module(context)


@then('the faster-whisper model used model size "{model_size}"')
def step_faster_whisper_used_model_size(context, model_size: str) -> None:
    _ = context
    faster_whisper_module = sys.modules.get("faster_whisper")
    assert faster_whisper_module is not None
    actual = getattr(faster_whisper_module, "last_model_size", None)
    assert actual == model_size


@then('the faster-whisper transcription used language "{language}"')
def step_faster_whisper_used_language(context, language: str) -> None:
    _ = context
    faster_whisper_module = sys.modules.get("faster_whisper")
    assert faster_whisper_module is not None
    actual = getattr(faster_whisper_module, "last_language", None)
    assert actual == language


@then('the faster-whisper transcription used beam size {beam_size:d}')
def step_faster_whisper_used_beam_size(context, beam_size: int) -> None:
    _ = context
    faster_whisper_module = sys.modules.get("faster_whisper")
    assert faster_whisper_module is not None
    actual = getattr(faster_whisper_module, "last_beam_size", None)
    assert actual == beam_size



"""
Unit tests for text utility tool call sequencing.
"""

from __future__ import annotations

import json
import unittest

from biblicus.ai.llm import ChatCompletionResult
from biblicus.ai.models import AiProvider, LlmClientConfig
from biblicus.text import annotate as annotate_module
from biblicus.text import link as link_module
from biblicus.text import redact as redact_module
from biblicus.text import tool_loop as tool_loop_module
from biblicus.text.prompts import (
    DEFAULT_ANNOTATE_SYSTEM_PROMPT,
    DEFAULT_LINK_SYSTEM_PROMPT,
    DEFAULT_REDACT_SYSTEM_PROMPT,
)
from biblicus.text.tool_loop import run_tool_loop


class FakeAnnotateCompletion:
    """
    Fake annotate completion response with str_replace tool calls.
    """

    def __call__(self, **_kwargs: object) -> ChatCompletionResult:
        """
        Return a synthetic annotate completion result.
        """
        tool_calls = [
            {
                "id": "call_1",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps(
                        {
                            "old_str": "Long passage",
                            "new_str": '<span label="entity">Long passage',
                        }
                    ),
                },
            },
            {
                "id": "call_2",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps(
                        {
                            "old_str": "annotation.",
                            "new_str": "annotation.</span>",
                        }
                    ),
                },
            },
            {
                "id": "call_3",
                "type": "function",
                "function": {"name": "done", "arguments": "{}"},
            },
        ]
        return ChatCompletionResult(text="", tool_calls=tool_calls)


class FakeRedactCompletion:
    """
    Fake redact completion response with str_replace tool calls.
    """

    def __call__(self, **_kwargs: object) -> ChatCompletionResult:
        """
        Return a synthetic redact completion result.
        """
        tool_calls = [
            {
                "id": "call_1",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps(
                        {
                            "old_str": "Sensitive data",
                            "new_str": '<span redact="pii">Sensitive data',
                        }
                    ),
                },
            },
            {
                "id": "call_2",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps({"old_str": "removed.", "new_str": "removed.</span>"}),
                },
            },
            {
                "id": "call_3",
                "type": "function",
                "function": {"name": "done", "arguments": "{}"},
            },
        ]
        return ChatCompletionResult(text="", tool_calls=tool_calls)


class FakeLinkCompletion:
    """
    Fake link completion response with str_replace tool calls.
    """

    def __call__(self, **_kwargs: object) -> ChatCompletionResult:
        """
        Return a synthetic link completion result.
        """
        tool_calls = [
            {
                "id": "call_1",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps(
                        {
                            "old_str": "Acme updated. ",
                            "new_str": '<span id="link_1">Acme updated. ',
                        }
                    ),
                },
            },
            {
                "id": "call_2",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps(
                        {"old_str": "Acme updated. ", "new_str": "Acme</span> updated. "}
                    ),
                },
            },
            {
                "id": "call_3",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps(
                        {
                            "old_str": "Acme updated.",
                            "new_str": '<span ref="link_1">Acme updated.',
                        }
                    ),
                },
            },
            {
                "id": "call_4",
                "type": "function",
                "function": {
                    "name": "str_replace",
                    "arguments": json.dumps(
                        {"old_str": "Acme updated.", "new_str": "Acme</span> updated."}
                    ),
                },
            },
            {
                "id": "call_5",
                "type": "function",
                "function": {"name": "done", "arguments": "{}"},
            },
        ]
        return ChatCompletionResult(text="", tool_calls=tool_calls)


class TestTextUtilityToolCalls(unittest.TestCase):
    """
    Unit tests for text utility tool call sequencing.
    """

    def test_text_annotate_uses_two_str_replace_calls(self) -> None:
        """
        Ensure annotate uses two str_replace tool calls.
        """
        text = "Long passage needing annotation."
        allowed_attributes = ["label"]
        system_prompt = annotate_module._render_system_prompt(
            DEFAULT_ANNOTATE_SYSTEM_PROMPT,
            allowed_attributes=allowed_attributes,
        )
        fake_completion = FakeAnnotateCompletion()
        original = tool_loop_module.chat_completion
        tool_loop_module.chat_completion = fake_completion
        try:
            result = run_tool_loop(
                text=text,
                client=LlmClientConfig(
                    provider=AiProvider.OPENAI,
                    model="gpt-4o-mini",
                    api_key="test-openai-key",
                    response_format="json_object",
                ),
                system_prompt=system_prompt,
                prompt_template="Return the requested text.",
                max_rounds=1,
                max_edits_per_round=5,
                apply_str_replace=annotate_module._apply_annotate_replace,
                validate_text=lambda current_text: annotate_module._validate_annotation_markup(
                    current_text, allowed_attributes
                ),
            )
        finally:
            tool_loop_module.chat_completion = original

        assistant_messages = [
            message for message in result.messages if message.get("role") == "assistant"
        ]
        tool_calls = assistant_messages[0].get("tool_calls") or []
        str_replace_calls = [
            call for call in tool_calls if call.get("function", {}).get("name") == "str_replace"
        ]
        self.assertEqual(len(str_replace_calls), 2)
        self.assertIn("<span", result.text)
        self.assertIn("</span>", result.text)

    def test_text_redact_uses_two_str_replace_calls(self) -> None:
        """
        Ensure redact uses two str_replace tool calls.
        """
        text = "Sensitive data should be removed."
        redaction_types = ["pii"]
        system_prompt = redact_module._render_system_prompt(
            DEFAULT_REDACT_SYSTEM_PROMPT,
            redaction_types=redaction_types,
        )
        fake_completion = FakeRedactCompletion()
        original = tool_loop_module.chat_completion
        tool_loop_module.chat_completion = fake_completion
        try:
            result = run_tool_loop(
                text=text,
                client=LlmClientConfig(
                    provider=AiProvider.OPENAI,
                    model="gpt-4o-mini",
                    api_key="test-openai-key",
                    response_format="json_object",
                ),
                system_prompt=system_prompt,
                prompt_template="Return the requested text.",
                max_rounds=1,
                max_edits_per_round=5,
                apply_str_replace=redact_module._apply_redact_replace,
                validate_text=lambda current_text: redact_module._validate_redaction_markup(
                    current_text, redaction_types
                ),
            )
        finally:
            tool_loop_module.chat_completion = original

        assistant_messages = [
            message for message in result.messages if message.get("role") == "assistant"
        ]
        tool_calls = assistant_messages[0].get("tool_calls") or []
        str_replace_calls = [
            call for call in tool_calls if call.get("function", {}).get("name") == "str_replace"
        ]
        self.assertEqual(len(str_replace_calls), 2)
        self.assertIn("<span", result.text)
        self.assertIn("</span>", result.text)

    def test_text_link_uses_four_str_replace_calls(self) -> None:
        """
        Ensure link uses four str_replace tool calls.
        """
        text = "Acme updated. Acme updated."
        system_prompt = link_module._render_system_prompt(
            DEFAULT_LINK_SYSTEM_PROMPT, id_prefix="link_"
        )
        fake_completion = FakeLinkCompletion()
        original = tool_loop_module.chat_completion
        tool_loop_module.chat_completion = fake_completion
        try:
            result = run_tool_loop(
                text=text,
                client=LlmClientConfig(
                    provider=AiProvider.OPENAI,
                    model="gpt-4o-mini",
                    api_key="test-openai-key",
                    response_format="json_object",
                ),
                system_prompt=system_prompt,
                prompt_template="Link repeated mentions.",
                max_rounds=1,
                max_edits_per_round=10,
                apply_str_replace=link_module._apply_link_replace,
                validate_text=lambda current_text: link_module._validate_link_markup(
                    current_text, "link_"
                ),
            )
        finally:
            tool_loop_module.chat_completion = original

        assistant_messages = [
            message for message in result.messages if message.get("role") == "assistant"
        ]
        tool_calls = assistant_messages[0].get("tool_calls") or []
        str_replace_calls = [
            call for call in tool_calls if call.get("function", {}).get("name") == "str_replace"
        ]
        self.assertEqual(len(str_replace_calls), 4)
        self.assertIn('id="link_1"', result.text)
        self.assertIn('ref="link_1"', result.text)


if __name__ == "__main__":
    unittest.main()



import importlib.util

import pytest

from biblicus.graph.extractors.dependency_relations import DependencyRelationsGraphExtractor
from biblicus.graph.extractors.ner_entities import NerEntitiesGraphExtractor
from biblicus.graph.extractors.simple_entities import SimpleEntitiesGraphExtractor
from biblicus.models import CatalogItem, GraphExtractionResult
from biblicus.corpus import Corpus


def _text_item():
    return CatalogItem(
        id="i",
        relpath="raw/a.txt",
        sha256="h",
        bytes=1,
        media_type="text/plain",
        title=None,
        tags=[],
        metadata={},
        created_at="now",
        source_uri="file://a",
    )


def test_dependency_relations_validates_config(tmp_path):
    extractor = DependencyRelationsGraphExtractor()
    item = _text_item()
    corpus = Corpus(tmp_path)
    if importlib.util.find_spec("spacy") is None:
        with pytest.raises(ValueError):
            extractor.extract_graph(
                corpus=corpus,
                item=item,
                extracted_text="alpha beta",
                config={"model": "en_core_web_sm", "min_entity_length": 2, "include_item_node": False},
            )
    else:
        res = extractor.extract_graph(
            corpus=corpus,
            item=item,
            extracted_text="alpha beta",
            config={"model": "en_core_web_sm", "min_entity_length": 2, "include_item_node": False},
        )
        assert isinstance(res, GraphExtractionResult)


def test_ner_entities_validates_config(tmp_path):
    extractor = NerEntitiesGraphExtractor()
    item = _text_item()
    corpus = Corpus(tmp_path)
    if importlib.util.find_spec("spacy") is None:
        with pytest.raises(ValueError):
            extractor.extract_graph(
                corpus=corpus,
                item=item,
                extracted_text="New York is great",
                config={"model": "en_core_web_sm", "min_entity_length": 2, "include_item_node": False},
            )
    else:
        res = extractor.extract_graph(
            corpus=corpus,
            item=item,
            extracted_text="New York is great",
            config={"model": "en_core_web_sm", "min_entity_length": 2, "include_item_node": False},
        )
        assert isinstance(res, GraphExtractionResult)


def test_simple_entities_validates_config(tmp_path):
    extractor = SimpleEntitiesGraphExtractor()
    item = _text_item()
    corpus = Corpus(tmp_path)
    res = extractor.extract_graph(
        corpus=corpus,
        item=item,
        extracted_text="Entity one",
        config={"model": "en_core_web_sm", "min_entity_length": 2, "include_item_node": False},
    )
    assert isinstance(res, GraphExtractionResult)



"""
Deepgram transcription transform extractor.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from ..corpus import Corpus
from ..models import CatalogItem, ExtractedText, ExtractionStageOutput
from .base import TextExtractor


class DeepgramTranscriptTransformConfig(BaseModel):
    """
    Configuration for transforming Deepgram structured metadata into text.

    :ivar source: Which Deepgram representation to render (transcript, utterances, or words).
    :vartype source: str
    :ivar channels: Optional list of channel indices to include.
    :vartype channels: list[int] or None
    :ivar speakers: Optional list of speaker indices to include.
    :vartype speakers: list[int] or None
    :ivar join_with: Separator used when joining utterances or words.
    :vartype join_with: str
    :ivar include_channel_labels: Whether to prefix text with channel labels.
    :vartype include_channel_labels: bool
    :ivar include_speaker_labels: Whether to prefix text with speaker labels.
    :vartype include_speaker_labels: bool
    """

    model_config = ConfigDict(extra="forbid")

    source: str = Field(default="transcript", min_length=1)
    channels: Optional[List[int]] = None
    speakers: Optional[List[int]] = None
    join_with: str = Field(default=" ", min_length=1)
    include_channel_labels: bool = False
    include_speaker_labels: bool = False


class DeepgramTranscriptTransformExtractor(TextExtractor):
    """
    Transform Deepgram structured metadata into text output.

    :ivar extractor_id: Extractor identifier.
    :vartype extractor_id: str
    """

    extractor_id = "deepgram-transform"

    def validate_config(self, config: Dict[str, Any]) -> BaseModel:
        """
        Validate transform configuration.

        :param config: Configuration mapping.
        :type config: dict[str, Any]
        :return: Parsed configuration.
        :rtype: DeepgramTranscriptTransformConfig
        """
        parsed = DeepgramTranscriptTransformConfig.model_validate(config)
        if parsed.source not in {"transcript", "utterances", "words"}:
            raise ValueError(
                "deepgram-transform source must be one of transcript, utterances, or words"
            )
        return parsed

    def extract_text(
        self,
        *,
        corpus: Corpus,
        item: CatalogItem,
        config: BaseModel,
        previous_extractions: List[ExtractionStageOutput],
    ) -> Optional[ExtractedText]:
        """
        Produce text from Deepgram structured metadata in prior stage outputs.

        :param corpus: Corpus containing the item bytes.
        :type corpus: Corpus
        :param item: Catalog item being processed.
        :type item: CatalogItem
        :param config: Parsed configuration model.
        :type config: DeepgramTranscriptTransformConfig
        :param previous_extractions: Prior stage outputs for this item within the pipeline.
        :type previous_extractions: list[biblicus.models.ExtractionStageOutput]
        :return: Extracted text payload or None when metadata is missing.
        :rtype: ExtractedText or None
        """
        _ = corpus
        if not item.media_type.startswith("audio/"):
            return None

        parsed = (
            config
            if isinstance(config, DeepgramTranscriptTransformConfig)
            else DeepgramTranscriptTransformConfig.model_validate(config)
        )

        deepgram_payload = _find_deepgram_payload(previous_extractions=previous_extractions)
        if deepgram_payload is None:
            raise ValueError("deepgram-transform requires Deepgram metadata from a prior stage")

        text = _render_deepgram_text(payload=deepgram_payload, config=parsed)
        return ExtractedText(text=text, producer_extractor_id=self.extractor_id)


def _find_deepgram_payload(
    *, previous_extractions: List[ExtractionStageOutput]
) -> Optional[Dict[str, Any]]:
    for output in reversed(previous_extractions):
        if not output.metadata:
            continue
        payload = output.metadata.get("deepgram")
        if isinstance(payload, dict):
            return payload
    return None


def _render_deepgram_text(
    *, payload: Dict[str, Any], config: DeepgramTranscriptTransformConfig
) -> str:
    results = payload.get("results") or {}
    channels = results.get("channels") or []
    channel_indices = (
        [index for index in range(len(channels))]
        if config.channels is None
        else list(config.channels)
    )
    chunks: List[str] = []
    for channel_index, channel in enumerate(channels):
        if channel_index not in channel_indices:
            continue
        alternatives = channel.get("alternatives") or []
        if not alternatives:
            continue
        alternative = alternatives[0] or {}
        if config.source == "transcript":
            transcript = str(alternative.get("transcript") or "").strip()
            if transcript:
                chunks.append(_prefix_with_labels(
                    transcript,
                    channel_index=channel_index,
                    speaker=None,
                    config=config,
                ))
            continue
        if config.source == "utterances":
            utterances = alternative.get("utterances") or results.get("utterances") or []
            utterance_texts: List[str] = []
            for entry in utterances:
                if not isinstance(entry, dict):
                    continue
                speaker = entry.get("speaker")
                channel_value = entry.get("channel")
                if config.channels is not None and channel_value is not None:
                    if int(channel_value) not in channel_indices:
                        continue
                if config.speakers is not None and speaker is not None:
                    if int(speaker) not in config.speakers:
                        continue
                text_value = str(entry.get("transcript") or entry.get("text") or "").strip()
                if not text_value:
                    continue
                utterance_texts.append(
                    _prefix_with_labels(
                        text_value,
                        channel_index=channel_index if channel_value is None else int(channel_value),
                        speaker=int(speaker) if speaker is not None else None,
                        config=config,
                    )
                )
            if utterance_texts:
                chunks.append(config.join_with.join(utterance_texts))
            continue
        words = alternative.get("words") or results.get("words") or []
        word_texts: List[str] = []
        current_speaker: Optional[int] = None
        current_channel: Optional[int] = channel_index
        for entry in words:
            if not isinstance(entry, dict):
                continue
            speaker = entry.get("speaker")
            channel_value = entry.get("channel")
            if config.channels is not None and channel_value is not None:
                if int(channel_value) not in channel_indices:
                    continue
            if config.speakers is not None and speaker is not None:
                if int(speaker) not in config.speakers:
                    continue
            word_value = str(entry.get("punctuated_word") or entry.get("word") or "").strip()
            if not word_value:
                continue
            if config.include_speaker_labels and speaker is not None:
                speaker_value = int(speaker)
                channel_value_int = int(channel_value) if channel_value is not None else channel_index
                if word_texts and (speaker_value != current_speaker or channel_value_int != current_channel):
                    chunks.append(
                        _prefix_with_labels(
                            config.join_with.join(word_texts),
                            channel_index=current_channel,
                            speaker=current_speaker,
                            config=config,
                        )
                    )
                    word_texts = []
                current_speaker = speaker_value
                current_channel = channel_value_int
            word_texts.append(word_value)
        if word_texts:
            chunks.append(
                _prefix_with_labels(
                    config.join_with.join(word_texts),
                    channel_index=current_channel if config.include_speaker_labels else channel_index,
                    speaker=current_speaker if config.include_speaker_labels else None,
                    config=config,
                )
            )
    return "\n".join(chunk for chunk in chunks if chunk).strip()


def _prefix_with_labels(
    text: str,
    *,
    channel_index: Optional[int],
    speaker: Optional[int],
    config: DeepgramTranscriptTransformConfig,
) -> str:
    labels: List[str] = []
    if config.include_channel_labels and channel_index is not None:
        labels.append(f"Channel {channel_index}")
    if config.include_speaker_labels and speaker is not None:
        labels.append(f"Speaker {speaker}")
    if not labels:
        return text
    return f"{', '.join(labels)}: {text}"



"""
Task planning utilities for Biblicus workflows.
"""

from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from .corpus import Corpus
from .extraction import build_extraction_snapshot, create_extraction_configuration_manifest
from .models import ExtractionSnapshotListEntry, RetrievalSnapshot
from .retrieval import create_configuration_manifest
from .retrievers import get_retriever

TASK_KIND_ALIASES = {
    "fetch": "load",
    "sync": "load",
    "build": "index",
    "run": "query",
}


class Task(BaseModel):
    """
    Planned task with dependency metadata.

    :ivar name: User-facing task name.
    :vartype name: str
    :ivar kind: Canonical task kind.
    :vartype kind: str
    :ivar target_type: Target type (corpus or retriever).
    :vartype target_type: str
    :ivar target_id: Identifier for the task target.
    :vartype target_id: str
    :ivar depends_on: Dependent tasks that must run first.
    :vartype depends_on: list[Task]
    :ivar status: Current task status (ready, complete, blocked).
    :vartype status: str
    :ivar reason: Optional reason when blocked.
    :vartype reason: str or None
    :ivar metadata: Task-specific configuration metadata.
    :vartype metadata: dict[str, Any]
    """

    model_config = ConfigDict(extra="forbid")

    name: str
    kind: str
    target_type: str
    target_id: str
    depends_on: List["Task"] = Field(default_factory=list)
    status: str = Field(default="ready")
    reason: Optional[str] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)


Task.model_rebuild()


class Plan(BaseModel):
    """
    Ordered task plan with dependency resolution results.

    :ivar tasks: Ordered list of tasks to execute.
    :vartype tasks: list[Task]
    :ivar root: Root task for the plan.
    :vartype root: Task
    :ivar status: Aggregate plan status (ready, complete, blocked).
    :vartype status: str
    :ivar missing_handlers: Task kinds without a registered handler.
    :vartype missing_handlers: list[str]
    """

    model_config = ConfigDict(extra="forbid")

    tasks: List[Task]
    root: Task
    status: str = "ready"
    missing_handlers: List[str] = Field(default_factory=list)

    def execute(
        self,
        *,
        mode: str = "prompt",
        handler_registry: Optional[Dict[str, Callable[[Task], Any]]] = None,
        prompt_handler: Optional[Callable[[Task], bool]] = None,
    ) -> List[Any]:
        """
        Execute the plan in dependency order.

        :param mode: Execution mode (prompt, auto, dry_run).
        :type mode: str
        :param handler_registry: Mapping of task kinds to callables.
        :type handler_registry: dict[str, Callable[[Task], Any]] or None
        :param prompt_handler: Callback to request permission for a task.
        :type prompt_handler: Callable[[Task], bool] or None
        :return: List of handler results.
        :rtype: list[Any]
        :raises RuntimeError: If a task is blocked or no handler exists.
        """
        if mode not in {"prompt", "auto", "dry_run"}:
            raise ValueError(f"Unsupported plan execution mode: {mode}")
        handler_registry = handler_registry or {}
        results: List[Any] = []
        self.missing_handlers = []

        for task in self.tasks:
            if task.status == "complete":
                continue
            if task.status == "blocked":
                raise RuntimeError(task.reason or f"Task '{task.name}' is blocked")
            if mode == "dry_run":
                continue
            if mode == "prompt":
                if prompt_handler is None or not prompt_handler(task):
                    raise RuntimeError(f"Task '{task.name}' declined by prompt handler")
            handler = handler_registry.get(task.kind)
            if handler is None:
                task.status = "blocked"
                task.reason = f"No handler registered for task kind '{task.kind}'"
                self.missing_handlers.append(task.kind)
                self.status = "blocked"
                raise RuntimeError(task.reason)
            results.append(handler(task))
            task.status = "complete"

        self.status = _plan_status(self.tasks)
        return results


def normalize_task_kind(name: str) -> str:
    """
    Normalize a task name into the canonical task kind.

    :param name: Task name or alias.
    :type name: str
    :return: Canonical task kind.
    :rtype: str
    """
    if not name:
        return name
    lower = name.strip().lower()
    return TASK_KIND_ALIASES.get(lower, lower)


def _default_pipeline_config() -> Dict[str, Any]:
    return {
        "stages": [
            {
                "extractor_id": "pass-through-text",
                "configuration": {},
            }
        ]
    }


def _pipeline_configuration_id(pipeline_config: Dict[str, Any]) -> str:
    manifest = create_extraction_configuration_manifest(
        extractor_id="pipeline",
        name="default",
        configuration=pipeline_config,
    )
    return manifest.configuration_id


def _catalog_generated_at(corpus: Corpus) -> str:
    return corpus.catalog_generated_at()


def _list_retrieval_snapshots(corpus: Corpus) -> List[RetrievalSnapshot]:
    snapshots: List[RetrievalSnapshot] = []
    if not corpus.retrieval_dir.is_dir():
        return snapshots
    for retriever_dir in sorted(corpus.retrieval_dir.iterdir()):
        if not retriever_dir.is_dir():
            continue
        for snapshot_dir in sorted(retriever_dir.iterdir()):
            manifest_path = snapshot_dir / "manifest.json"
            if not manifest_path.is_file():
                continue
            try:
                snapshots.append(corpus.load_snapshot(snapshot_dir.name))
            except Exception:
                continue
    return snapshots


def _find_retrieval_snapshot(
    *,
    corpus: Corpus,
    retriever_id: str,
    configuration_id: str,
    catalog_generated_at: str,
) -> Optional[RetrievalSnapshot]:
    for snapshot in _list_retrieval_snapshots(corpus):
        if snapshot.configuration.retriever_id != retriever_id:
            continue
        if snapshot.configuration.configuration_id != configuration_id:
            continue
        if snapshot.catalog_generated_at != catalog_generated_at:
            continue
        return snapshot
    return None


def _find_extraction_snapshot(
    *,
    corpus: Corpus,
    configuration_id: str,
    catalog_generated_at: str,
) -> Optional[ExtractionSnapshotListEntry]:
    entries = corpus.list_extraction_snapshots(extractor_id="pipeline")
    for entry in entries:
        if entry.configuration_id != configuration_id:
            continue
        if entry.catalog_generated_at != catalog_generated_at:
            continue
        return entry
    return None


def _build_task(
    *,
    name: str,
    kind: str,
    target_type: str,
    target_id: str,
    depends_on: Optional[List[Task]] = None,
    status: str = "ready",
    reason: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
) -> Task:
    return Task(
        name=name,
        kind=kind,
        target_type=target_type,
        target_id=target_id,
        depends_on=list(depends_on or []),
        status=status,
        reason=reason,
        metadata=dict(metadata or {}),
    )


def _flatten_tasks(task: Task, seen: Optional[set[int]] = None) -> List[Task]:
    seen = seen or set()
    tasks: List[Task] = []
    for child in task.depends_on:
        tasks.extend(_flatten_tasks(child, seen))
    if id(task) not in seen:
        tasks.append(task)
        seen.add(id(task))
    return tasks


def _plan_status(tasks: List[Task]) -> str:
    if any(task.status == "blocked" for task in tasks):
        return "blocked"
    if all(task.status == "complete" for task in tasks):
        return "complete"
    return "ready"


def build_default_handler_registry(
    corpus: Corpus,
) -> Dict[str, Callable[[Task], Any]]:
    """
    Build default task handlers for extraction and indexing.

    :param corpus: Corpus to operate on.
    :type corpus: Corpus
    :return: Mapping of task kinds to handler functions.
    :rtype: dict[str, Callable[[Task], Any]]
    """

    def _handle_extract(task: Task) -> Any:
        pipeline_config = task.metadata.get("pipeline") or _default_pipeline_config()
        force = bool(task.metadata.get("force", False))
        max_workers = int(task.metadata.get("max_workers", 1))
        return build_extraction_snapshot(
            corpus,
            extractor_id="pipeline",
            configuration_name="default",
            configuration=pipeline_config,
            force=force,
            max_workers=max_workers,
        )

    def _handle_index(task: Task) -> Any:
        retriever = get_retriever(task.target_id)
        index_config = task.metadata.get("index_config") or {}
        return retriever.build_snapshot(
            corpus, configuration_name="index", configuration=index_config
        )

    return {
        "extract": _handle_extract,
        "index": _handle_index,
    }


def build_plan_for_extract(
    corpus: Corpus,
    *,
    pipeline_config: Optional[Dict[str, Any]] = None,
    load_handler_available: bool = False,
    force: bool = False,
    max_workers: int = 1,
) -> Plan:
    """
    Build a dependency plan for corpus extraction.

    :param corpus: Corpus to plan against.
    :type corpus: Corpus
    :param pipeline_config: Optional extraction pipeline configuration.
    :type pipeline_config: dict[str, Any] or None
    :param load_handler_available: Whether a load handler is available.
    :type load_handler_available: bool
    :param force: Whether to force extraction even when snapshots exist.
    :type force: bool
    :param max_workers: Maximum number of concurrent extraction workers.
    :type max_workers: int
    :return: Planned task graph for extraction.
    :rtype: Plan
    """
    pipeline_config = pipeline_config or _default_pipeline_config()
    catalog_generated_at = _catalog_generated_at(corpus)
    configuration_id = _pipeline_configuration_id(pipeline_config)
    snapshot = _find_extraction_snapshot(
        corpus=corpus,
        configuration_id=configuration_id,
        catalog_generated_at=catalog_generated_at,
    )
    has_items = corpus.has_items()

    extract_status = "complete" if snapshot and not force else "ready"
    extract_task = _build_task(
        name="extract",
        kind="extract",
        target_type="corpus",
        target_id=corpus.uri,
        status=extract_status,
        metadata={
            "pipeline": pipeline_config,
            "force": force,
            "max_workers": max_workers,
        },
    )

    if not has_items:
        if load_handler_available:
            load_task = _build_task(
                name="load",
                kind="load",
                target_type="corpus",
                target_id=corpus.uri,
                status="ready",
            )
            extract_task.depends_on.append(load_task)
        else:
            extract_task.status = "complete"
            extract_task.reason = None

    tasks = _flatten_tasks(extract_task)
    return Plan(tasks=tasks, root=extract_task, status=_plan_status(tasks))


def build_plan_for_index(
    corpus: Corpus,
    retriever_id: str,
    *,
    pipeline_config: Optional[Dict[str, Any]] = None,
    index_config: Optional[Dict[str, Any]] = None,
    load_handler_available: bool = False,
) -> Plan:
    """
    Build a dependency plan for retriever indexing.

    :param corpus: Corpus to plan against.
    :type corpus: Corpus
    :param retriever_id: Retriever identifier.
    :type retriever_id: str
    :param pipeline_config: Optional extraction pipeline configuration.
    :type pipeline_config: dict[str, Any] or None
    :param index_config: Optional retriever index configuration.
    :type index_config: dict[str, Any] or None
    :param load_handler_available: Whether a load handler is available.
    :type load_handler_available: bool
    :return: Planned task graph for indexing.
    :rtype: Plan
    """
    pipeline_config = pipeline_config or _default_pipeline_config()
    index_config = index_config or {}
    catalog_generated_at = _catalog_generated_at(corpus)

    configuration_manifest = create_configuration_manifest(
        retriever_id=retriever_id,
        name="index",
        configuration=index_config,
    )
    snapshot = _find_retrieval_snapshot(
        corpus=corpus,
        retriever_id=retriever_id,
        configuration_id=configuration_manifest.configuration_id,
        catalog_generated_at=catalog_generated_at,
    )
    index_status = "complete" if snapshot else "ready"
    index_task = _build_task(
        name="index",
        kind="index",
        target_type="retriever",
        target_id=retriever_id,
        status=index_status,
        metadata={"index_config": index_config},
    )

    extract_plan = build_plan_for_extract(
        corpus,
        pipeline_config=pipeline_config,
        load_handler_available=load_handler_available,
    )
    index_task.depends_on.append(extract_plan.root)
    if extract_plan.root.status == "blocked":
        index_task.status = "blocked"
        index_task.reason = extract_plan.root.reason
    tasks = _flatten_tasks(index_task)
    return Plan(tasks=tasks, root=index_task, status=_plan_status(tasks))


def build_plan_for_query(
    corpus: Corpus,
    retriever_id: str,
    *,
    pipeline_config: Optional[Dict[str, Any]] = None,
    index_config: Optional[Dict[str, Any]] = None,
    load_handler_available: bool = False,
) -> Plan:
    """
    Build a dependency plan for retrieval queries.

    :param corpus: Corpus to plan against.
    :type corpus: Corpus
    :param retriever_id: Retriever identifier.
    :type retriever_id: str
    :param pipeline_config: Optional extraction pipeline configuration.
    :type pipeline_config: dict[str, Any] or None
    :param index_config: Optional retriever index configuration.
    :type index_config: dict[str, Any] or None
    :param load_handler_available: Whether a load handler is available.
    :type load_handler_available: bool
    :return: Planned task graph for querying.
    :rtype: Plan
    """
    index_plan = build_plan_for_index(
        corpus,
        retriever_id,
        pipeline_config=pipeline_config,
        index_config=index_config,
        load_handler_available=load_handler_available,
    )
    query_task = _build_task(
        name="query",
        kind="query",
        target_type="retriever",
        target_id=retriever_id,
        status="ready",
    )
    if index_plan.root.status != "complete":
        query_task.depends_on.append(index_plan.root)
    if index_plan.root.status == "blocked":
        query_task.status = "blocked"
        query_task.reason = index_plan.root.reason
    tasks = _flatten_tasks(query_task)
    return Plan(tasks=tasks, root=query_task, status=_plan_status(tasks))


def build_plan_for_load(corpus: Corpus, *, handler_available: bool = False) -> Plan:
    """
    Build a dependency plan for corpus loading.

    :param corpus: Corpus to plan against.
    :type corpus: Corpus
    :param handler_available: Whether a load handler is available.
    :type handler_available: bool
    :return: Planned task graph for loading.
    :rtype: Plan
    """
    has_items = corpus.has_items()
    status = "complete" if has_items else "ready" if handler_available else "blocked"
    reason = None
    if status == "blocked":
        reason = "Corpus has no items and no load handler is available"
    load_task = _build_task(
        name="load",
        kind="load",
        target_type="corpus",
        target_id=corpus.uri,
        status=status,
        reason=reason,
    )
    tasks = [load_task]
    return Plan(tasks=tasks, root=load_task, status=_plan_status(tasks))



"""
Selection extractor that chooses the longest available text from previous pipeline outputs.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict

from ..models import CatalogItem, ExtractedText, ExtractionStageOutput
from .base import TextExtractor


class SelectLongestTextExtractorConfig(BaseModel):
    """
    Configuration for the longest text selection extractor.

    Version zero does not expose configuration for this extractor.
    """

    model_config = ConfigDict(extra="forbid")


class SelectLongestTextExtractor(TextExtractor):
    """
    Extractor plugin that selects the longest text from previous pipeline outputs.

    This extractor does not attempt to score semantic quality. It is a deterministic
    selection policy for cases where multiple stages can produce usable text for the
    same item.

    The selection rules are:

    - If any prior extracted texts are non-empty after stripping whitespace, choose the one
      with the greatest stripped character count.
    - Ties are broken by earliest pipeline stage index.
    - If no prior extracted texts are usable but prior extracted texts exist, select the
      earliest extracted text even if it is empty.

    :ivar extractor_id: Extractor identifier.
    :vartype extractor_id: str
    """

    extractor_id = "select-longest-text"

    def validate_config(self, config: Dict[str, Any]) -> BaseModel:
        """
        Validate selection extractor configuration.

        :param config: Configuration mapping.
        :type config: dict[str, Any]
        :return: Parsed configuration.
        :rtype: SelectLongestTextExtractorConfig
        """
        return SelectLongestTextExtractorConfig.model_validate(config)

    def extract_text(
        self,
        *,
        corpus,
        item: CatalogItem,
        config: BaseModel,
        previous_extractions: List[ExtractionStageOutput],
    ) -> Optional[ExtractedText]:
        """
        Select the longest extracted text from previous pipeline outputs.

        :param corpus: Corpus containing the item bytes.
        :type corpus: Corpus
        :param item: Catalog item being processed.
        :type item: CatalogItem
        :param config: Parsed configuration model.
        :type config: SelectLongestTextExtractorConfig
        :param previous_extractions: Prior stage outputs for this item within the pipeline.
        :type previous_extractions: list[biblicus.models.ExtractionStageOutput]
        :return: Selected extracted text payload or None when no prior outputs exist.
        :rtype: ExtractedText or None
        """
        _ = corpus
        _ = item
        _ = config

        extracted_candidates = [entry for entry in previous_extractions if entry.text is not None]
        if not extracted_candidates:
            return None

        usable_candidates = [entry for entry in extracted_candidates if entry.text.strip()]
        if usable_candidates:
            candidate = max(usable_candidates, key=lambda entry: len(entry.text.strip()))
            ties = [
                entry
                for entry in usable_candidates
                if len(entry.text.strip()) == len(candidate.text.strip())
            ]
            candidate = min(ties, key=lambda entry: int(entry.stage_index))
        else:
            candidate = min(extracted_candidates, key=lambda entry: int(entry.stage_index))

        producer = candidate.producer_extractor_id or candidate.extractor_id
        return ExtractedText(
            text=candidate.text or "",
            producer_extractor_id=producer,
            source_stage_index=candidate.stage_index,
        )



from behave import then, when

from biblicus.sources import _namespaced_filename, _sanitize_filename_component


def _normalize_value(value: str) -> str:
    return "" if value == "<empty>" else value


@when('I sanitize the filename "{name}"')
def step_sanitize_filename(context, name: str) -> None:
    context.sanitized_filename = _sanitize_filename_component(_normalize_value(name))


@when(
    'I build a namespaced filename from source uri "{source_uri}" and fallback "{fallback}" with media type "{media_type}"'
)
def step_namespaced_filename(context, source_uri: str, fallback: str, media_type: str) -> None:
    source = _normalize_value(source_uri) or None
    fallback_name = _normalize_value(fallback) or None
    context.namespaced_filename = _namespaced_filename(
        source_uri=source,
        fallback_name=fallback_name,
        media_type=media_type,
    )


@then('the sanitized filename equals "{expected}"')
def step_sanitized_filename_equals(context, expected: str) -> None:
    assert context.sanitized_filename == expected


@then('the namespaced filename equals "{expected}"')
def step_namespaced_filename_equals(context, expected: str) -> None:
    assert context.namespaced_filename == expected



"""
Default prompts for Biblicus text utilities.
"""

from __future__ import annotations

DEFAULT_PROMPT_TEMPLATE = "Return the requested text."

DEFAULT_EXTRACT_SYSTEM_PROMPT = (
    "You are a virtual file editor. Use the available tools to edit the text.\n"
    "Interpret the word 'return' in the user's request as: wrap the returned text with "
    "<span>...</span> in-place in the current text.\n\n"
    "Use the str_replace tool to insert <span>...</span> tags and the done tool when finished.\n"
    "For long spans, insert <span> and </span> using separate str_replace calls. "
    "For short spans (a few words), it is acceptable to insert both tags in one call.\n"
    "When finished, call done. Do NOT return JSON in the assistant message.\n\n"
    "Rules:\n"
    "- Use str_replace only.\n"
    "- old_str must match exactly once in the current text.\n"
    "- When choosing old_str, copy the exact substring (including punctuation/case) from the current text.\n"
    "- old_str and new_str must be non-empty strings.\n"
    "- new_str must be identical to old_str with only <span> and/or </span> inserted.\n"
    "- Do not include <span> or </span> inside old_str.\n"
    "- Do not insert nested spans.\n"
    "- If a tool call fails due to non-unique old_str, retry with a longer unique old_str.\n"
    "- If a tool call fails, read the error and keep editing. Do not call done until spans are inserted.\n"
    "- Do not delete, reorder, paraphrase, or label text.\n\n"
    "Current text:\n---\n{text}\n---\n"
)

DEFAULT_SLICE_SYSTEM_PROMPT = (
    "You are a virtual file editor. Use the available tools to edit the text.\n"
    "Interpret the word 'return' in the user's request as: insert <slice/> markers in-place in the current text.\n\n"
    "Use the str_replace tool to insert <slice/> markers and the done tool when finished.\n"
    "When finished, call done. Do NOT return JSON in the assistant message.\n\n"
    "Rules:\n"
    "- Use str_replace only.\n"
    "- old_str must match exactly once in the current text.\n"
    "- old_str and new_str must be non-empty strings.\n"
    "- new_str must be identical to old_str with only <slice/> inserted.\n"
    "- Do not include <slice/> inside old_str or new_str.\n"
    "- If a tool call fails due to non-unique old_str, retry with a longer unique old_str.\n"
    "- If a tool call fails, read the error and keep editing. Do not call done until markers are inserted.\n"
    "- Do not delete, reorder, or paraphrase text.\n\n"
    "Current text:\n---\n{text}\n---\n"
)

DEFAULT_ANNOTATE_SYSTEM_PROMPT = (
    "You are a virtual file editor. Use the available tools to edit the text.\n"
    "Interpret the user's request as: wrap the requested text with "
    '<span ATTRIBUTE="VALUE">...</span> in-place in the current text.\n'
    "Each span must include exactly one attribute from: {{ allowed_attributes }}.\n\n"
    "Use the str_replace tool to insert span tags and the done tool when finished.\n"
    "For long spans, insert the opening and closing tags using separate str_replace calls. "
    "For short spans (a few words), it is acceptable to insert both tags in one call.\n"
    "When finished, call done. Do NOT return JSON in the assistant message.\n\n"
    "Rules:\n"
    "- Use str_replace only.\n"
    "- old_str must match exactly once in the current text.\n"
    "- old_str and new_str must be non-empty strings.\n"
    "- new_str must be identical to old_str with only <span ...> and/or </span> inserted.\n"
    "- Do not include <span or </span> inside old_str.\n"
    "- Do not insert nested spans.\n"
    "- Do not wrap text that is already inside a span; spans must never overlap.\n"
    "- If a name appears inside an existing span, leave it alone and wrap only bare text.\n"
    "- If a tool call fails due to non-unique old_str, retry with a longer unique old_str.\n"
    "- If a tool call fails, read the error and keep editing. Do not call done until spans are inserted.\n"
    "- Do not delete, reorder, paraphrase, or label text beyond the span attributes.\n\n"
    "Current text:\n---\n{text}\n---\n"
)

DEFAULT_LINK_SYSTEM_PROMPT = (
    "You are a virtual file editor. Use the available tools to edit the text.\n"
    "Interpret the word 'return' in the user's request as: wrap the returned text with "
    '<span ATTRIBUTE="VALUE">...</span> in-place in the current text.\n'
    "Each span must include exactly one attribute: id for first mentions and ref for repeats.\n"
    "Id values must start with '{{ id_prefix }}'.\n\n"
    "Linking rules:\n"
    "- For each distinct name or entity, assign exactly one id on its first occurrence.\n"
    "- Use ref on every subsequent occurrence of the same name or entity.\n"
    "- Wrap only the repeated name or entity text itself (no extra surrounding words).\n"
    "- Reuse the same id/ref value for identical names; do not create multiple ids for the same name.\n"
    "- Use ids in order of first appearance ({{ id_prefix }}1, {{ id_prefix }}2, ...).\n"
    "- Do not call done until every repeated name or entity in the text is wrapped.\n"
    "- If a name appears multiple times, there must be one id and refs for every later occurrence.\n\n"
    "Use the str_replace tool to insert span tags and the done tool when finished.\n"
    "For long spans, insert the opening and closing tags using separate str_replace calls. "
    "For short spans (a few words), it is acceptable to insert both tags in one call.\n"
    "When finished, call done. Do NOT return JSON in the assistant message.\n\n"
    "Rules:\n"
    "- Use str_replace only.\n"
    "- old_str must match exactly once in the current text.\n"
    "- old_str and new_str must be non-empty strings.\n"
    "- new_str must be identical to old_str with only <span ...> and/or </span> inserted.\n"
    "- Do not include <span or </span> inside old_str.\n"
    "- Do not insert nested spans.\n"
    "- If a tool call fails due to non-unique old_str, retry with a longer unique old_str.\n"
    "- If a tool call fails, read the error and keep editing. Do not call done until spans are inserted.\n"
    "- Do not delete, reorder, or paraphrase text.\n\n"
    "Current text:\n---\n{text}\n---\n"
)

DEFAULT_REDACT_SYSTEM_PROMPT = (
    "You are a virtual file editor. Use the available tools to edit the text.\n"
    "Interpret the word 'return' in the user's request as: wrap the returned text with "
    "<span>...</span> in-place in the current text.\n"
    "If redaction types are provided, use a redact attribute with one of: {{ redaction_types }}.\n\n"
    "Use the str_replace tool to insert span tags and the done tool when finished.\n"
    "For long spans, insert the opening and closing tags using separate str_replace calls. "
    "For short spans (a few words), it is acceptable to insert both tags in one call.\n"
    "When finished, call done. Do NOT return JSON in the assistant message.\n\n"
    "Rules:\n"
    "- Use str_replace only.\n"
    "- old_str must match exactly once in the current text.\n"
    "- old_str and new_str must be non-empty strings.\n"
    "- new_str must be identical to old_str with only <span ...> and/or </span> inserted.\n"
    "- Do not include <span or </span> inside old_str.\n"
    "- Do not insert nested spans.\n"
    "- If a tool call fails due to non-unique old_str, retry with a longer unique old_str.\n"
    "- If a tool call fails, read the error and keep editing. Do not call done until spans are inserted.\n"
    "- Do not delete, reorder, or paraphrase text.\n\n"
    "Current text:\n---\n{text}\n---\n"
)



import json
from pathlib import Path

from biblicus.analysis import markov


def test_load_topic_modeling_report_invalid_returns_none(tmp_path: Path, capsys):
    run_dir = tmp_path
    # valid JSON but incompatible schema to trigger validation fallback
    (run_dir / "topic_modeling.json").write_text('{"bad": "data"}', encoding="utf-8")
    assert markov._load_topic_modeling_report(run_dir=run_dir) is None


def test_load_llm_observation_cache_handles_bad_entries(tmp_path: Path):
    path = tmp_path / "cache.json"
    payload = {
        "segments": ["bad", {"segment_index": "x"}, {"segment_index": 2, "value": "ok"}],
    }
    path.write_text(json.dumps(payload), encoding="utf-8")
    cache = markov._load_llm_observation_cache(path)
    assert cache == {2: {"segment_index": 2, "value": "ok"}}


def test_load_llm_observation_cache_missing_file():
    assert markov._load_llm_observation_cache(Path("/nonexistent/noway")) == {}



import sys
import types
from pathlib import Path

import pytest

from biblicus.analysis.topic_modeling import (
    TopicModelingDocument,
    _apply_entity_removal,
    _apply_llm_extraction,
)
from biblicus.analysis.models import (
    TopicModelingEntityRemovalConfig,
    TopicModelingLlmExtractionConfig,
    TopicModelingLlmExtractionMethod,
)


def test_entity_removal_provider_not_spacy_validates():
    config = TopicModelingEntityRemovalConfig(enabled=True, provider="spacy")
    # ensure no exception when provider is correct
    assert config.provider == "spacy"


def test_entity_removal_logs_progress_large(monkeypatch, capsys):
    docs = [TopicModelingDocument(str(i), "s", "text") for i in range(1200)]

    class FakeDoc:
        def __init__(self, text):
            self.text = text
            self.ents = []

    def fake_load(model):  # noqa: ARG001
        return lambda text: FakeDoc(text)

    fake_spacy = types.SimpleNamespace(load=fake_load)
    monkeypatch.setitem(sys.modules, "spacy", fake_spacy)

    config = TopicModelingEntityRemovalConfig(enabled=True)
    report, processed = _apply_entity_removal(documents=docs, config=config, cache_path=None)
    assert report.status.value == "complete"
    assert len(processed) == 1200
    captured = capsys.readouterr()
    assert "topic-modeling" in captured.err


def test_llm_itemized_empty_raises(monkeypatch):
    docs = [TopicModelingDocument("d1", "s1", "alpha"), TopicModelingDocument("d2", "s2", "beta")]
    config = TopicModelingLlmExtractionConfig(
        enabled=True,
        client={"provider": "mock", "model": "m"},
        prompt_template="{text}",
        method=TopicModelingLlmExtractionMethod.ITEMIZE,
    )
    monkeypatch.setattr(
        "biblicus.analysis.topic_modeling.generate_completion",
        lambda **_: "\n\n",
    )
    with pytest.raises(ValueError):
        _apply_llm_extraction(documents=docs, config=config)
